with Messages;
use Messages;
with Ada.Text_IO,Ada.Integer_Text_IO;
use Ada.Text_IO,Ada.Integer_Text_IO;


pragma Elaborate_All (Messages);

package body Beltrami_Ops is

   pragma Suppress(Storage_Check);


   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;

   pragma Inline (Imin,Imax);
   pragma Suppress(Storage_Check);


   function T_TestFunc(R_0: Radius; P, Q: Integer; C:  Comp) return Comp is
      R1,R2,R,RR,CC: Comp;
   begin
      if P<=0 or Q<=0  then
         Message("Beltrami_Ops.T_TestFunc",Parameter_Error);
      end if;
      R:=Comp_One;
      for I in 1..2*(Q+1) loop
         R:=R_0*R;
      end loop;
      R1:=Comp_One;
      for I in 1..(Q+1) loop
         R1:=Conj(C)*R1;
      end loop;
      R2:=Comp_One;
      for I in 1..(P-1) loop
         R2:=C*R2;
      end loop;
      RR:=R1*R2;
      CC:=Comp_One;
      if (Q-P+1)>0 then
         for I in 1..(Q+1-P) loop
            CC:=CC*Inv(C);
         end loop;
      else
         for I in 1..(P-Q-1) loop
            CC:=CC*C ;
         end loop;
      end if;
      if AbsVal(C)=Zero then
         if (P-Q=2) then
            return Neg(R)/Rep(Q+1);
         else
            return Comp_Zero;
         end if;
      elsif AbsVal(C)<R_0 then
         if P-Q-1>=0 then
            return Rep(Q-P+1)/Rep(Q+1)*R/C*CC+Rep(P)/Rep(Q+1)*RR;
         else
            return Rep(P)/Rep(Q+1)*RR;
         end if;
      elsif AbsVal(C)=R_0 then
         if P-Q-1>=0 then
            return Half*Rep(Q-P+1)/Rep(Q+1)*R/C*CC+Half*Rep(P)/Rep(Q+1)*RR;
         else
            return Half*Rep(P-Q-1)/Rep(Q+1)*R/C*CC+Half*Rep(P)/Rep(Q+1)*RR;
         end if;
      else
         if Q+1-P>0 then
            return Rep(-Q+P-1)/Rep(Q+1)*R*CC/C;
         else
            return Comp_Zero;
         end if;
      end if;
   end T_TestFunc;


   function P_TestFunc(R_0: Radius; P, Q: Integer; C:  Comp) return Comp is
      R1,R2,R,RR,CC: Comp;
   begin
      if P<=0 or Q<=0  then
         Message("Beltrami_Ops.P_TestFunc",Parameter_Error);
      end if;
      R:=Comp_One;
      for I in 1..2*(Q+1) loop
         R:=R_0*R;
      end loop;
      R1:=Comp_One;
      for I in 1..(Q+1) loop
         R1:=Conj(C)*R1;
      end loop;
      R2:=Comp_One;
      for I in 1..P loop
         R2:=C*R2;
      end loop;
      RR:=R1*R2;
      CC:=Comp_One;
      if (Q-P+1)>0 then
         for I in 1..(Q+1-P) loop
            CC:=CC*Inv(C);
         end loop;
      else
         for I in 1..(P-Q-1) loop
            CC:=CC*C ;
         end loop;
      end if;
      if AbsVal(C)=Zero then
         if (P-Q=1) then
            return Neg(R)/Rep(Q+1);
         else
            return Comp_Zero;
         end if;
      elsif AbsVal(C)<R_0 then
         if P-Q-1>=0 then
            return One/Rep(Q+1)*(Neg(R)*CC+RR);
         else
            return RR/Rep(Q+1);
         end if;
      elsif AbsVal(C)=R_0 then
         if P-Q-1>=0 then
            return Half/Rep(Q+1)*(Neg(R)*CC+RR);
         else
            return Half/Rep(Q+1)*(R*CC+RR);
         end if;
      else
         if Q+1-P>0 then
            return R*CC/Rep(Q+1);
         else
            return Comp_Zero;
         end if;
      end if;
   end P_TestFunc;


   procedure Glassman (A, B, C:   in  Integer; Data_Vector:   in out Vector ;Inverse_Transform  :   in  Boolean:=False   )  is
      Temp:  Vector(1..Data_Vector'length);
      Counter:  Integer := Data_Vector'first;
      JC:  Integer := 0;
      Del, Omega, Sum:  Comp;
      Angle:  Rep;
      C_Plus_1:  Integer := C + 1; --=3 in the standard case
   begin
      Temp(1..Data_Vector'length) := Data_Vector;
      Angle:= Two*Pi/(Rep(A * C));
      Del:= Complecs(Cos(Angle),-Sin(Angle));
      Omega:=Comp_One;
      if Inverse_Transform then
         Del := Conj(Del);
      end if;
      for IC in 1..C loop --executed two times
         for IA in 1..A loop
            for IB in 1..B loop
               Sum      := Temp((((IA - 1)*C + (C-1))*B) + IB);--Temp[(N/2^k)*(2*(IA - 1)+1)+IB]
               for JCR in 2..C loop  -- executed once in the standard case, JCR=2
                  JC := C_Plus_1 - JCR; -- No need to add C + 1 each
                                        -- time through loop
                                        -- JC=1 in the standard case
                  Sum := Temp((((IA - 1) * C + (JC - 1)) * B)
                              + IB) + (Omega  *  Sum); --Temp[(N/2^k)*2*(IA-1)+IB]+W^(IC*IA)*Temp[(N/2^k)*(2*(IA - 1)+1)+IB]
               end loop; -- JCR
               Data_Vector(Counter) := Sum;
               Counter := Counter +   1;
            end loop;  -- IB
            Omega := Del  *  Omega;
         end loop; -- IA
      end loop; -- IC
   end Glassman;


   procedure FFT (FFT_Data:  in out Vector; Inverse_Transform: in   Boolean:=False) is
      A: Integer   := 1;
      B: Integer   := FFT_Data'length;
      C: Integer   := 1;
   begin -- FFT
      while (B  >  1)  loop   -- define the integers A, B, and C
         A := C  *  A;       --  such that A * B * C = FFT_Data'length
         C := 2;
         while (B mod C) /= 0 loop
            C := C + 1;
         end loop;
         B := B/C;     -- B = 1 causes exit from while loop
         Glassman (A,B,C, FFT_Data, Inverse_Transform);
      end loop;
      if not Inverse_Transform  then
         for i in FFT_Data'range loop
            FFT_Data(i) := FFT_Data(i)/Rep(FFT_Data'length);
         end loop;
      end if;
   end FFT;


   procedure FT(H: in Matrix; FH: out Matrix; Inverse_Transform: in   Boolean:=False )
   is
      -- The first row in the input is the value at the center if Inverse_Transform:=False; and zeros except for H(First,0), otherwise
      -- The i-th row in the input is the values at 2^N angles on the the circle of radius R_i if Inverse_Transform:=False; and Fourier components otherwise
      P: Vector(H'Range(2));
      Centre: Comp;
      M: Rep;
   begin
      if H'First(2)/=-H'Length(2)/2 then
         Message("Beltrami_Ops.FT",Index_Error);
      end if;
      Centre:= H(H'First(1),0);
      if Inverse_Transform then
         for J in H'Range(2) loop
            FH(H'First(1),J ):=Centre;
         end loop;
      else
         for J in H'Range(2) loop
            if J/=0 then
               FH(H'First(1),J ):=Comp_Zero;
            else
               FH(H'First(1),J ):=Centre;
            end if;
         end loop;
      end if;
      for I in H'First(1)+1..H'Last(1) loop
         --tilde{U}_n=U_{n-N/2}(-1)^n
         M:=One;
         for J in H'Range(2) loop
            P(J):=M*H(I,J);
            M:=-M;
         end loop;
         FFT(P,Inverse_Transform);
         --tilde{U}_m=U_{m-N/2}(-1)^m
         M:=One;
         for J in P'Range loop
            FH(I,J):=M*P(J);
            M:=-M;
         end loop;
      end loop;
   end FT;


   procedure FT(H: in out Matrix; Inverse_Transform: in   Boolean:=False )  is
      -- The first row in the input is the value at the center if Inverse_Transform:=False; and zeros except for H(First,0), otherwise
      -- The i-th row in the input is the values at 2^N angles on the the circle of radius R_i if Inverse_Transform:=False; and Fourier components otherwise
      A,B,C: Integer;
      Temp: Vector(H'Range(2));
      M,Div: Rep;
      Com: Comp;
   begin
      if H'First(2)/=-H'Length(2)/2 then
         Message("RG_Ops.FT",Index_Error);
      end if;
      if not Inverse_Transform  then
         Div := Rep(H'Length(2));
      else
         Div:=One;
      end if;
      for I in H'First(1)+1..H'Last(1) loop
         --tilde{U}_n=U_{n-N/2}(-1)^n
         --tilde{U}_m=U_{m-N/2}(-1)^m
         A:= 1;
         B:= H'Length(2);
         C:= 1;
         M:=One;
         for J in H'Range(2) loop
            Temp(J):=M*H(I,J);
            M:=-M;
         end loop;
         while (B  >  1)  loop   -- define the integers A, B, and C
            A := C  *  A;       --  such that A * B * C = FFT_Data'length
            C := 2;
            while (B mod C) /= 0 loop
               C := C + 1;
            end loop;
            B:=B/C;
            Glassman (A,B,C,Temp,Inverse_Transform);
         end loop;
         for J in Temp'Range loop
            H(I,J):=Temp(J)/Div;
            Div:=-Div;
         end loop;
      end loop;
      Com:=H(H'First(1),0);
      if Inverse_Transform then --point values
         for J in H'Range(2) loop
            H(H'First(1),J):=Com;
         end loop;
      else
         for J in H'Range(2) loop
            H(H'First(1),J):=Comp_Zero;
         end loop;
         H(H'First(1),0):=Com;
      end if;
   end FT;


   procedure FT1(H: in out Matrix; Inverse_Transform: in   Boolean:=False )  is
      -- The first row in the input is the value at the center if Inverse_Transform:=False; and zeros except for H(First,0), otherwise
      -- The i-th row in the input is the values at 2^N angles on the the circle of radius R_i if Inverse_Transform:=False; and Fourier components otherwise
      A,B,C: Integer;
      Temp: Vector(H'Range(2));
      M,Div: Rep;
   begin
      if H'First(2)/=-H'Length(2)/2 then
         Message("Beltrami_Ops.FT",Index_Error);
      end if;

      if not Inverse_Transform  then
         Div := Rep(H'Length(2));
      else
         Div:=One;
      end if;
      for I in H'Range(1) loop
         --tilde{U}_n=U_{n-N/2}(-1)^n
         --tilde{U}_m=U_{m-N/2}(-1)^m
         A:= 1;
         B:= H'Length(2);
         C:= 1;
         M:=One;
         for J in H'Range(2) loop
            Temp(J):=M*H(I,J);
            M:=-M;
         end loop;
         while (B  >  1)  loop   -- define the integers A, B, and C
            A := C  *  A;       --  such that A * B * C = FFT_Data'length
            C := 2;
            while (B mod C) /= 0 loop
               C := C + 1;
            end loop;
            B:=B/C;  -- B = 1 causes exit from while loop
            Glassman (A,B,C,Temp,Inverse_Transform);
         end loop;
         for J in Temp'Range loop
            H(I,J):=Temp(J)/Div;
            Div:=-Div;
         end loop;
      end loop;
   end FT1;


   function Value(V: in Vector; C: in Comp) return Comp is
      --V(Arg(C))
      CC,E,Ex: Comp;
   begin
      if Arg(C)=Zero then
         E:=Comp_One;
      else
         E:=C*Inv(  Complecs(AbsVal(C),Zero)  );
      end if;
      Ex:=Comp_One;
      for I in V'First..-1 loop
         Ex:=Ex*Inv(E);
      end loop;
      CC:=Comp_Zero;
      for I in V'Range loop
         CC:=CC+V(I)*Ex;
         Ex:=Ex*E;
      end loop;
      return CC;
   end Value;


   function Value(Rad: in RepArray; Centre: in Comp; F: in Matrix; C: in Comp) return Comp is
      --assumes a Fourier series around Centre
      V1,V2: Vector(F'Range(2));
      A,Dell: Rep;
   begin
      A:=AbsVal(C-Centre);
      if A>=Rad(Rad'Last) then
         for J in F'Range(2) loop
            V1(J):=F(F'Last(1),J);
         end loop;
         return Value(V1,C-Centre);
      elsif A=Zero then
         for J in F'Range(2) loop
            V1(J):=F(F'First(1),J);
         end loop;
         return Value(V1,C-Centre);
      else
         for I in Rad'Range loop
            if Rad(I)>=A then
               for J in F'Range(2) loop
                  V1(J):=F(I-1,J);
                  V2(J):=F(I,J);
               end loop;
               Dell:= (Rad(I)-A)/(Rad(I)-Rad(I-1));
               return      Dell*Value(V1,C-Centre)+(One-Dell)*Value(V2,C-Centre);
            end if;
         end loop;
      end if;
   end Value;


   procedure T_Transform(Lg,R: in RepArray; S: in RealMatrix; H: in out Matrix; Text_Output: in Boolean:=False)  is
      --used for solving the Beltrami equation on a compact disk of radius R_0,
      -- negative I's coresspond to the inside of the disk,
      -- positive I's - to the otside
      pragma Suppress(Storage_Check);
      C1,C2,Den: Rep;
   begin
      if H'Length(2)<4 then
         Message("Beltrami_Ops.T_Transform",Index_Error);
      end if;
      if H'First(1)/=R'First or H'Last(1)/=R'Last or H'Last(1)<-1  then
         Message("Beltrami_Ops.T_Transform",Index_Error);
      end if;
      if Text_Output then
         New_Line;
         Put("Angular grid dimensions:=(");ShortPut(H'First(2));Put(",");ShortPut(H'Last(2));Put(")");
         New_Line;
         Put("Angular grid size:="); Txt_Put(Current_Output,Two*Pi/Rep(H'Last(2)-H'First(2)));
         New_Line;
         Put("Radial grid dimensions:=(");ShortPut(H'First(1));Put(",");ShortPut(H'Last(1));Put(")");
         New_Line;
         Put("Radii:");New_Line;
         Show(R);
      end if;
      --NEGATIVE J's
      for J in H'First(2)..-1 loop
         H(H'First(1),J):=Comp_Zero;
         --I=-N+1
         H(H'First(1)+1,J):=(One-Two*Rep(J+1)/Rep(J-1))*H(H'First(1)+1,J+2)+Two*Rep(J+1)/Rep(J*(J-1))*H(H'First(1),J+2);
         ---N+1<I<0
         for I in H'First(1)+2..-1 loop
            Den:=S(I-1,1)-One;
            C1:=Two*Rep(J+1)/Rep(J-1)*(S(I-1,-J)*S(I-1,1)-One)/Den;
            C2:=Two*Rep(J+1)/Rep(J)*(S(I-1,-J)-One)/Den;
            H(I,J):=S(I-1,-J)*H(I-1,J)+(C1-C2-S(I-1,-J))*H(I-1,J+2)+(One-C1+S(I-1,1)*C2)*H(I,J+2);
         end loop;
         if H'Last(1)>=0 then
            --I=0
            Den:=S(-1,1)-One;
            C1:=Two*Rep(J+1)/Rep(J-1)*(S(-1,-J)*S(-1,1)-One)/Den;
            C2:=Two*Rep(J+1)/Rep(J)*(S(-1,-J)-One)/Den;
            H(0,J):=S(-1,-J)*H(-1,J)+(C1-C2-S(-1,-J))*H(-1,J+2);
         end if;
         --I>0
         for I in 1..H'Last(1) loop
            H(I,J):=S(I-1,-J)*H(I-1,J);
         end loop;
      end loop;
      --NONNEGATIVE J's
      --J=0;1
      --I>=0
      for I in reverse 0..H'Last(1) loop
         H(I,0):=Comp_Zero;
         H(I,1):=Comp_Zero;
      end loop;
      --I=-1
      if H'Last(1)=-1 then
         H(-1,0):=H(-1,2);
         H(-1,1):=H(-1,3);
      elsif H'Last(1)>-1 then
         Den:=S(-1,1)-One;
         H(-1,0):=Three*H(-1,2)-Two*Lg(-1)*H(-1,2)/Den;
         H(-1,1):=Four*S(-1,1)*Lg(-1)*H(-1,3)/Den-Three*H(-1,3);
      else
         New_Line;
         Message("Beltrami_Ops.T_Transform",Index_Error);
      end if;
      --I<-1
      for I in reverse H'First(1)+1..-2 loop
         Den:=S(I,1)-One;
         H(I,0):=H(I+1,0)+Three*(H(I,2)-H(I+1,2))-Two*Lg(I)*(H(I,2)-S(I,1)*H(I+1,2))/Den;
         H(I,1):=S(I,1)*H(I+1,1)-Three*(H(I,3)-S(I,1)*H(I+1,3))+Four*S(I,1)*Lg(I)*(H(I,3)-H(I+1,3))/Den;
      end loop;
      if H(H'First(1),2)/=Comp_Zero then
         New_Line;
         Message("RG_Ops",Division_By_Zero);
      end if;
      --I=-N
      H(H'First(1),0):=H(H'First(1)+1,0)-Three*H(H'First(1)+1,2);
      H(H'First(1),1):=Comp_Zero;
      --J>=2
      for J in 2..H'Last(2)-2 loop
         --I>=0
         for I in reverse 0..H'Last(1) loop
            H(I,J):=Comp_Zero;
         end loop;
         --I=-1
         if H'Last(1)=-1 then
            H(-1,J):=H(-1,J+2);
         elsif H'Last(1)>-1 then
            Den:=S(-1,1)-One;
            C1:=Two*Rep(J+1)/Rep(J-1)*(S(-1,J)-S(-1,1))/Den;
            C2:=Two*Rep(J+1)/Rep(J)*(S(-1,J)-One)/Den;
            H(-1,J):=(One+C1-C2)*H(-1,J+2);
         else
            New_Line;
            Message("Beltrami_Ops.T_Transform",Index_Error);
         end if;
         --I<=-1
         for I in reverse H'First(1)+1..-2 loop
            Den:=S(I,1)-One;
            C1:=Two*Rep(J+1)/Rep(J-1)*((S(I,J)-S(I,1))/Den);
            C2:=Two*Rep(J+1)/Rep(J)*((S(I,J)-One)/Den);
            H(I,J):=S(I,J)*H(I+1,J)+(One+C1-C2)*H(I,J+2)+(S(I,1)*C2-C1-S(I,J))*H(I+1,J+2);
         end loop;
         --I=-N
         H(H'First(1),J):=Comp_Zero;
      end loop;

      for J in H'Last(2)-1..H'Last(2) loop
         for I in H'Range(1) loop
            H(I,J):=Comp_Zero;
         end loop;
      end loop;
   end T_Transform;


   procedure T_Transform(R: in RepArray; H: in out Matrix; Text_Output: in Boolean:=False)  is
      --used for solving the Beltrami equation on a compact disk of radius R_0,
      -- negative I's coresspond to the inside of the disk,
      -- positive I's - to the otside
      pragma Suppress(Storage_Check);
      P: Matrix(H'Range(1),0..1);
      C1,C2,C3,Eps: Rep;
      JJ,J: Integer;
   begin
      if H'First(1)>=0 or H'Last(1)<0 then
         Message("RG_Ops.T_Transform",Index_Error);
      end if;
      if H'Length(2)<4 then
         Message("RG_Ops.T_Transform",Index_Error);
      end if;
      if Text_Output then
         New_Line;
         Put("Angular grid dimensions:=(");ShortPut(H'First(2));Put(",");ShortPut(H'Last(2));Put(")");
         New_Line;
         Put("Angular grid size:="); Txt_Put(Current_Output,Two*Pi/Rep(H'Last(2)-H'First(2)));
         New_Line;
         Put("Radial grid dimensions:=(");ShortPut(H'First(1));Put(",");ShortPut(H'Last(1));Put(")");
         New_Line;
         Put("Radii:");New_Line;
         Show(R);
      end if;
      for I in P'Range(1) loop
         P(I,0):=H(I,0);
         P(I,1):=H(I,1);
      end loop;
      --NONNEGATIVE J's
      --I>=0
      for I in reverse 0..H'Last(1) loop  -- values for J>=0 outside of the circle and on the boundary
         H(I,0):=Comp_Zero;
         H(I,1):=Comp_Zero;
      end loop;
      --I=-1;
      Eps:=R(0)-R(-1);
      C1:=Four/Eps;
      C2:=R(0)/R(-1);
      H(-1,0):=H(-1,2)-((Two/Eps)*Log(C2)*(R(0)*H(-1,2)-R(-1)*H(0,2)) + Two*(H(0,2)-H(-1,2)));
      H(-1,1):=H(-1,3)-(  C1*( (C2-One)*(R(0)*H(-1,3)-R(-1)*H(0,3))+R(0)*Log(C2)*(H(0,3)-H(-1,3)) ))/C2;
      --I<-1
      for I in reverse H'First(1)+1..-2 loop -- values for J>=0 inside  the circle minus zero, R(I)>0
         Eps:=R(I+1)-R(I);
         C1:=Four/Eps;
         C2:=R(I+1)/R(I);
         H(I,0):=H(I+1,0)-H(I+1,2)+H(I,2)-((Two/Eps)*Log(C2)*(R(I+1)*H(I,2)-R(I)*H(I+1,2)) + Two*(H(I+1,2)-H(I,2)));
         H(I,1):=(H(I+1,1)-H(I+1,3)-(C1*((C2-One)*(R(I+1)*H(I,3)-R(I)*H(I+1,3))+R(I+1)*Log(C2)*(H(I+1,3)-H(I,3)) )))/C2+H(I,3);
      end loop;
      if H(H'First(1),2)/=Comp_Zero then
         New_Line;
         Message("RG_Ops",Division_By_Zero);
      end if;
      H(H'First(1),0):=H(H'First(1)+1,0)-Three*H(H'First(1)+1,2);
      H(H'First(1),1):=Comp_Zero;
      --J>=2
      for K in -H'First(2)+2..H'Length(2)-1 loop
         J:=K+H'First(2);
         JJ:=((K+2) mod H'Length(2))+H'First(2);
         H(H'First(1),J):=Comp_Zero;
         --I>=0
         for I in reverse 0..H'Last(1) loop  -- values for J>=0 outside of the circle and on the boundary
            H(I,J):=Comp_Zero;
         end loop;
         --I=-1
         Eps:=R(0)-R(-1);
         C1:=Two*Rep(J+1)/(Rep(J)*Eps);
         C2:=Two*Rep(J+1)/(Rep(J-1)*Eps);
         C3:=(R(0)/R(-1))**J;
         --I<-1
         for I in reverse H'First(1)+1..-2 loop -- values for J>=0 inside  the circle minus zero, R(I)>0
            Eps:=R(I+1)-R(I);
            C1:=Two*Rep(J+1)/(Rep(J)*Eps);
            C2:=Two*Rep(J+1)/(Rep(J-1)*Eps);
            C3:=(R(I+1)/R(I))**J;
            H(I,J):=(H(I+1,J)-H(I+1,JJ)-(C1*(C3-One)*(R(I+1)*H(I,JJ)-R(I)*H(I+1,JJ))+C2*(C3-R(I+1)/R(I))*R(I)*(H(I+1,JJ)-H(I,JJ))))/C3+H(I,JJ);
         end loop;
      end loop;
      --NEGATIVE J's
      for K in 0..-H'First(2)-3 loop
         J:=K+H'First(2);
         H(H'First(1),J):=Comp_Zero;
         --I=H'First(1)+1
         Eps:=R(H'First(1)+1)-R(H'First(1));
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         H(H'First(1)+1,J):=H(H'First(1)+1,J+2)+C1*R(H'First(1)+1)*H(H'First(1),J+2)+C2*R(H'First(1)+1)*(H(H'First(1)+1,J+2)-H(H'First(1),J+2));
         --I<0
         for I in H'First(1)+2..-1 loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*( H(I-1,J)-H(I-1,J+2))+H(I,J+2)+C1*(One-C3)*( R(I)*H(I-1,J+2)-R(I-1)*H(I,J+2) )+C2*(R(I)/R(I-1)-C3)*R(I-1)*(H(I,J+2)-H(I-1,J+2));
         end loop;
         --I=0
         Eps:=R(0)-R(-1);
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         C3:=(R(-1)/R(0))**(abs(J));
         H(0,J):=C3*(H(-1,J)-H(-1,J+2))+C1*(One-C3)*( R(0)*H(-1,J+2)-R(-1)*H(0,J+2))  + C2*(R(0)/R(-1)-C3)*R(-1)*( H(0,J+2)-H(-1,J+2));
         --I>0
         for I in 1..H'Last(1) loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*H(I-1,J)+C1*(One-C3)*( R(I)*H(I-1,J+2)-R(I-1)*H(I,J+2))+C2*(R(I)/R(I-1)-C3)*R(I-1)*( H(I,J+2)-H(I-1,J+2));
         end loop;
      end loop;
      for K in  -H'First(2)-2 .. -H'First(2)-1 loop
         J:=K+H'First(2);
         H(H'First(1),J):=Comp_Zero;
         --I=H'First(1)+1
         Eps:=R(H'First(1)+1)-R(H'First(1));
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         H(H'First(1)+1,J):=P(H'First(1)+1,J+2)+C1*R(H'First(1)+1)*P(H'First(1),J+2)+C2*R(H'First(1)+1)*(P(H'First(1)+1,J+2)-P(H'First(1),J+2));
         --I<0
         for I in H'First(1)+2..-1 loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*( H(I-1,J)-P(I-1,J+2))+P(I,J+2)+C1*(One-C3)*( R(I)*P(I-1,J+2)-R(I-1)*P(I,J+2) )+C2*(R(I)/R(I-1)-C3)*R(I-1)*(P(I,J+2)-P(I-1,J+2));
         end loop;
         --I=0
         Eps:=R(0)-R(-1);
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         C3:=(R(-1)/R(0))**(abs(J));
         H(0,J):=C3*(H(-1,J)-P(-1,J+2))+C1*(One-C3)*( R(0)*P(-1,J+2)-R(-1)*P(0,J+2))+C2*(R(0)/R(-1)-C3)*R(-1)*( P(0,J+2)-P(-1,J+2));
         --I>0
         for I in 1..H'Last(1) loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*H(I-1,J)+C1*(One-C3)*( R(I)*P(I-1,J+2)-R(I-1)*P(I,J+2))+C2*(R(I)/R(I-1)-C3)*R(I-1)*( P(I,J+2)-P(I-1,J+2));
         end loop;
      end loop;
   end T_Transform;


   procedure TT_Transform(R: in RepArray; H: in out Matrix; Text_Output: in Boolean:=False)  is
      --used for solving the Beltrami equation on a compact disk of radius R_0,
      -- negative I's coresspond to the inside of the disk,
      -- positive I's - to the otside
      pragma Suppress(Storage_Check);
      P: Matrix(H'Range(1),0..1);
      C1,C2,C3,Eps: Rep;
      JJ,J: Integer;
   begin
      if H'First(1)>=0 or H'Last(1)<0 then
         Message("RG_Ops.T_Transform",Index_Error);
      end if;
      if H'Length(2)<4 then
         Message("RG_Ops.T_Transform",Index_Error);
      end if;
      if Text_Output then
         New_Line;
         Put("Angular grid dimensions:=(");ShortPut(H'First(2));Put(",");ShortPut(H'Last(2));Put(")");
         New_Line;
         Put("Angular grid size:="); Txt_Put(Current_Output,Two*Pi/Rep(H'Last(2)-H'First(2)));
         New_Line;
         Put("Radial grid dimensions:=(");ShortPut(H'First(1));Put(",");ShortPut(H'Last(1));Put(")");
         New_Line;
         Put("Radii:");New_Line;
         Show(R);
      end if;
      for I in P'Range(1) loop
         P(I,0):=H(I,0);
         P(I,1):=H(I,1);
      end loop;
      --NONNEGATIVE J's
      --I>=0
      for I in reverse 0..H'Last(1) loop  -- values for J>=0 outside of the circle and on the boundary
         H(I,0):=Comp_Zero;
         H(I,1):=Comp_Zero;
      end loop;
      --I=-1;
      Eps:=R(0)-R(-1);
      C1:=Four/Eps;
      C2:=R(0)/R(-1);
      H(-1,0):=H(-1,2)-((Two/Eps)*Log(C2)*(R(0)*H(-1,2)-R(-1)*H(0,2)) + Two*(H(0,2)-H(-1,2)));
      H(-1,1):=H(-1,3)-(  C1*( (C2-One)*(R(0)*H(-1,3)-R(-1)*H(0,3))+R(0)*Log(C2)*(H(0,3)-H(-1,3)) ))/C2;
      --I<-1
      for I in reverse H'First(1)+1..-2 loop -- values for J>=0 inside  the circle minus zero, R(I)>0
         Eps:=R(I+1)-R(I);
         C1:=Four/Eps;
         C2:=R(I+1)/R(I);
         H(I,0):=H(I+1,0)-H(I+1,2)+H(I,2)-((Two/Eps)*Log(C2)*(R(I+1)*H(I,2)-R(I)*H(I+1,2)) + Two*(H(I+1,2)-H(I,2)));
         H(I,1):=(H(I+1,1)-H(I+1,3)-(C1*((C2-One)*(R(I+1)*H(I,3)-R(I)*H(I+1,3))+R(I+1)*Log(C2)*(H(I+1,3)-H(I,3)) )))/C2+H(I,3);
      end loop;
      if H(H'First(1),2)/=Comp_Zero then
         New_Line;
         Message("RG_Ops",Division_By_Zero);
      end if;
      H(H'First(1),0):=H(H'First(1)+1,0)-Three*H(H'First(1)+1,2);
      H(H'First(1),1):=Comp_Zero;
      --J>=2
      for J in 2..H'Last(2)-2 loop
         JJ:=J+2;
         H(H'First(1),J):=Comp_Zero;
         --I>=0
         for I in reverse 0..H'Last(1) loop  -- values for J>=0 outside of the circle and on the boundary
            H(I,J):=Comp_Zero;
         end loop;
         --I=-1
         Eps:=R(0)-R(-1);
         C1:=Two*Rep(J+1)/(Rep(J)*Eps);
         C2:=Two*Rep(J+1)/(Rep(J-1)*Eps);
         C3:=(R(0)/R(-1))**J;
         H(-1,J):=H(-1,JJ)-(C1*(C3-One)*(R(0)*H(-1,JJ)-R(-1)*H(0,JJ))+C2*(C3-R(0)/R(-1))*R(-1)*(H(0,JJ)-H(-1,JJ)))/C3;
         --I<-1
         for I in reverse H'First(1)+1..-2 loop -- values for J>=0 inside  the circle minus zero, R(I)>0
            Eps:=R(I+1)-R(I);
            C1:=Two*Rep(J+1)/(Rep(J)*Eps);
            C2:=Two*Rep(J+1)/(Rep(J-1)*Eps);
            C3:=(R(I+1)/R(I))**J;
            H(I,J):=(H(I+1,J)-H(I+1,JJ)-(C1*(C3-One)*(R(I+1)*H(I,JJ)-R(I)*H(I+1,JJ))+C2*(C3-R(I+1)/R(I))*R(I)*(H(I+1,JJ)-H(I,JJ))))/C3+H(I,JJ);
         end loop;
      end loop;
      for J in H'Last(2)-1..H'Last(2) loop
         H(H'First(1),J):=Comp_Zero;
         for I in reverse H'Range(1) loop
            H(I,J):=Comp_Zero;
         end loop;
      end loop;
      --NEGATIVE J's
      for K in 0..-H'First(2)-3 loop
         J:=K+H'First(2);
         H(H'First(1),J):=Comp_Zero;
         --I=H'First(1)+1
         Eps:=R(H'First(1)+1)-R(H'First(1));
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         H(H'First(1)+1,J):=H(H'First(1)+1,J+2)+C1*R(H'First(1)+1)*H(H'First(1),J+2)+C2*R(H'First(1)+1)*(H(H'First(1)+1,J+2)-H(H'First(1),J+2));
         --I<0
         for I in H'First(1)+2..-1 loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*( H(I-1,J)-H(I-1,J+2))+H(I,J+2)+C1*(One-C3)*( R(I)*H(I-1,J+2)-R(I-1)*H(I,J+2) )+C2*(R(I)/R(I-1)-C3)*R(I-1)*(H(I,J+2)-H(I-1,J+2));
         end loop;
         --I=0
         Eps:=R(0)-R(-1);
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         C3:=(R(-1)/R(0))**(abs(J));
         H(0,J):=C3*(H(-1,J)-H(-1,J+2))+C1*(One-C3)*( R(0)*H(-1,J+2)-R(-1)*H(0,J+2))  + C2*(R(0)/R(-1)-C3)*R(-1)*( H(0,J+2)-H(-1,J+2));
         --I>0
         for I in 1..H'Last(1) loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*H(I-1,J)+C1*(One-C3)*( R(I)*H(I-1,J+2)-R(I-1)*H(I,J+2))+C2*(R(I)/R(I-1)-C3)*R(I-1)*( H(I,J+2)-H(I-1,J+2));
         end loop;
      end loop;
      for K in  -H'First(2)-2 .. -H'First(2)-1 loop
         J:=K+H'First(2);
         H(H'First(1),J):=Comp_Zero;
         --I=H'First(1)+1
         Eps:=R(H'First(1)+1)-R(H'First(1));
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         H(H'First(1)+1,J):=P(H'First(1)+1,J+2)+C1*R(H'First(1)+1)*P(H'First(1),J+2)+C2*R(H'First(1)+1)*(P(H'First(1)+1,J+2)-P(H'First(1),J+2));
         --I<0
         for I in H'First(1)+2..-1 loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*( H(I-1,J)-P(I-1,J+2))+P(I,J+2)+C1*(One-C3)*( R(I)*P(I-1,J+2)-R(I-1)*P(I,J+2) )+C2*(R(I)/R(I-1)-C3)*R(I-1)*(P(I,J+2)-P(I-1,J+2));
         end loop;
         --I=0
         Eps:=R(0)-R(-1);
         C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
         C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
         C3:=(R(-1)/R(0))**(abs(J));
         H(0,J):=C3*(H(-1,J)-P(-1,J+2))+C1*(One-C3)*( R(0)*P(-1,J+2)-R(-1)*P(0,J+2))+C2*(R(0)/R(-1)-C3)*R(-1)*( P(0,J+2)-P(-1,J+2));
         --I>0
         for I in 1..H'Last(1) loop
            Eps:=R(I)-R(I-1);
            C1:=Two*Rep(J+1)/(Rep(abs(J))*Eps);
            C2:=Two*Rep(J+1)/(Rep(abs(J)+1)*Eps);
            C3:=(R(I-1)/R(I))**(abs(J));
            H(I,J):=C3*H(I-1,J)+C1*(One-C3)*( R(I)*P(I-1,J+2)-R(I-1)*P(I,J+2))+C2*(R(I)/R(I-1)-C3)*R(I-1)*( P(I,J+2)-P(I-1,J+2));
         end loop;
      end loop;
   end TT_Transform;


   procedure P_Transform(Rad: in RepArray; H: in out Matrix; Text_Output: in Boolean:=False)  is
      --    pragma Suppress(Storage_Check);
      C: Comp;
      V,P: Vector(H'Range(1));
      Eps: Rep;
      JJ: Integer;
   begin
      if H'First(1)>=0 or H'Last(1)<0 then
         Message("P_Transform.RG_Ops",Index_Error);
      end if;
      if H'Length(2)<4 then
         Message("P_Transform.RG_Ops",Index_Error);
      end if;
      if Text_Output then
         New_Line;
         Put("Angular grid dimensions:=(");ShortPut(H'First(2));Put(",");ShortPut(H'Last(2));Put(")");
         New_Line;
         Put("Angular grid size:="); Txt_Put(Current_Output,Two*Pi/Rep(H'Last(2)-H'First(2)));
         New_Line;
         Put("Radial grid dimensions:=(");ShortPut(H'First(1));Put(",");ShortPut(H'Last(1));Put(")");
         New_Line;
         Put("Radii:");New_Line;
         Show(Rad);
      end if;
      for I in P'Range loop
         P(I):=H(I,0);
      end loop;

      for I in V'Range loop
         V(I):=H(I,H'First(2));
      end loop;
      --J=0
      for I in 0..H'Last(1) loop
         H(I,0):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
      end loop;
      for I in reverse H'First(1)+1..-1 loop
         Eps:=Rad(I+1)-Rad(I);
         C:=Two*(Rad(I+1)*H(I,1)-Rad(I)*H(I+1,1)) +(Sqr(Rad(I+1))-Sqr(Rad(I)))*(H(I+1,1)-H(I,1))/Eps;
         H(I,0):=H(I+1,0)-C;
      end loop;
      Eps:=Rad(H'First(1)+1)-Rad(H'First(1));
      C:=Two*(Rad(H'First(1)+1)*H(H'First(1),1)-Rad(H'First(1))*H(H'First(1)+1,1)) +(Sqr(Rad(H'First(1)+1))-Sqr(Rad(H'First(1))))*(H(H'First(1)+1,1)-H(H'First(1),1))/Eps;
      H(H'First(1),0):=H(H'First(1)+1,0)-C;
      --J=1
      for I in 0..H'Last(1) loop
         H(I,1):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
      end loop;
      for I in reverse H'First(1)+1..-1 loop
         Eps:=Rad(I+1)-Rad(I);
         C:=(Two*Rad(I))*Log(Rad(I+1)/Rad(I))*(Rad(I+1)*H(I,2)-Rad(I)*H(I+1,2))/Eps+Two*Rad(I)*(H(I+1,2)-H(I,2));
         H(I,1):=(Rad(I)/Rad(I+1))*H(I+1,1)-C;
      end loop;
      H(H'First(1),1):=Comp_Zero;
      --J=2
      for I in 0..H'Last(1) loop
         H(I,2):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
      end loop;
      for I in reverse H'First(1)+1..-1 loop
         Eps:=Rad(I+1)-Rad(I);
         C:=(Two*Rad(I))*(One-Rad(I)/Rad(I+1))*(Rad(I+1)*H(I,3)-Rad(I)*H(I+1,3))/Eps+Two*Sqr(Rad(I))*Log(Rad(I+1)/Rad(I))*(H(I+1,3)-H(I,3))/Eps;
         H(I,2):=((Rad(I)/Rad(I+1))**2)*H(I+1,2)-C;
      end loop;
      H(H'First(1),2):=Comp_Zero;
      --J>=3
      for J in 3..H'Last(2) loop
         if J+1>H'Last(2) then
            JJ:=H'First(2)+((J+1) mod H'Last(2))-1;
         else JJ:=J+1;
         end if;
         for I in 0..H'Last(1) loop
            H(I,J):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
         end loop;
         for I in reverse H'First(1)+1..-1 loop
            Eps:=Rad(I+1)-Rad(I);
            C:=(Two*Rad(I))/Rep(J-1)*(One-(Rad(I)/Rad(I+1))**(J-1))*(Rad(I+1)*H(I,JJ)-Rad(I)*H(I+1,JJ))/Eps+(Two*Sqr(Rad(I)))/Rep(J-2)*(One-(Rad(I)/Rad(I+1))**(J-2))*(H(I+1,JJ)-H(I,JJ))/Eps;
            H(I,J):=((Rad(I)/Rad(I+1))**J)*H(I+1,J)-C;
         end loop;
         H(H'First(1),J):=Comp_Zero;
      end loop;
      --J<=-2
      for J in H'First(2)..-2 loop
         H(H'First(1),J):=Comp_Zero;--value at zero
         for I in H'First(1)+1..H'Last(1) loop
            Eps:=Rad(I)-Rad(I-1);
            C:=(Two*Rad(I))/Rep(abs(J)+1)*(One-(Rad(I-1)/Rad(I))**(abs(J)+1))*(Rad(I)*H(I-1,J+1)-Rad(I-1)*H(I,J+1))/Eps+(Two*Sqr(Rad(I)))/Rep(abs(J)+2)*(One-(Rad(I-1)/Rad(I))**(abs(J)+2))*(H(I,J+1)-H(I-1,J+1))/Eps;
            if I=H'First(1)+1 then
               H(I,J):=C;
            else
               H(I,J):=((Rad(I-1)/Rad(I))**(abs(J)))*H(I-1,J)+C;
            end if;
         end loop;
      end loop;
      --J=-1
      H(H'First(1),-1):=Comp_Zero;--value at zero
      for I in H'First(1)+1..H'Last(1) loop
         Eps:=Rad(I)-Rad(I-1);
         C:=Rad(I)*(One-(Rad(I-1)/Rad(I))**2)*(Rad(I)*P(I-1)-Rad(I-1)*P(I))/Eps+(Two*Sqr(Rad(I)))/Three*(One-(Rad(I-1)/Rad(I))**3)*(P(I)-P(I-1))/Eps;
         if I=H'First(1)+1 then
            H(I,-1):=C;
         else
            H(I,-1):=(Rad(I-1)/Rad(I))*H(I-1,-1)+C;
         end if;
      end loop;
   end P_Transform;


   procedure PP_Transform(Rad: in RepArray; H: in out Matrix; Text_Output: in Boolean:=False)  is
      --    pragma Suppress(Storage_Check);
      C: Comp;
      V,P: Vector(H'Range(1));
      Eps: Rep;
      JJ: Integer;
   begin
      if H'First(1)>=0 or H'Last(1)<0 then
         Message("P_Transform.RG_Ops",Index_Error);
      end if;
      if H'Length(2)<4 then
         Message("P_Transform.RG_Ops",Index_Error);
      end if;
      if Text_Output then
         New_Line;
         Put("Angular grid dimensions:=(");ShortPut(H'First(2));Put(",");ShortPut(H'Last(2));Put(")");
         New_Line;
         Put("Angular grid size:="); Txt_Put(Current_Output,Two*Pi/Rep(H'Last(2)-H'First(2)));
         New_Line;
         Put("Radial grid dimensions:=(");ShortPut(H'First(1));Put(",");ShortPut(H'Last(1));Put(")");
         New_Line;
         Put("Radii:");New_Line;
         Show(Rad);
      end if;
      for I in P'Range loop
         P(I):=H(I,0);
      end loop;
      for I in V'Range loop
         V(I):=H(I,H'First(2));
      end loop;
      --J=0
      for I in 0..H'Last(1) loop
         H(I,0):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
      end loop;
      for I in reverse H'First(1)+1..-1 loop
         Eps:=Rad(I+1)-Rad(I);
         C:=Two*(Rad(I+1)*H(I,1)-Rad(I)*H(I+1,1)) +(Sqr(Rad(I+1))-Sqr(Rad(I)))*(H(I+1,1)-H(I,1))/Eps;
         H(I,0):=H(I+1,0)-C;
      end loop;
      Eps:=Rad(H'First(1)+1)-Rad(H'First(1));
      C:=Two*(Rad(H'First(1)+1)*H(H'First(1),1)-Rad(H'First(1))*H(H'First(1)+1,1)) +(Sqr(Rad(H'First(1)+1))-Sqr(Rad(H'First(1))))*(H(H'First(1)+1,1)-H(H'First(1),1))/Eps;
      H(H'First(1),0):=H(H'First(1)+1,0)-C;
      --J=1
      for I in 0..H'Last(1) loop
         H(I,1):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
      end loop;
      for I in reverse H'First(1)+1..-1 loop
         Eps:=Rad(I+1)-Rad(I);
         C:=(Two*Rad(I))*Log(Rad(I+1)/Rad(I))*(Rad(I+1)*H(I,2)-Rad(I)*H(I+1,2))/Eps+Two*Rad(I)*(H(I+1,2)-H(I,2));
         H(I,1):=(Rad(I)/Rad(I+1))*H(I+1,1)-C;
      end loop;
      H(H'First(1),1):=Comp_Zero;
      --J=2
      for I in 0..H'Last(1) loop
         H(I,2):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
      end loop;
      for I in reverse H'First(1)+1..-1 loop
         Eps:=Rad(I+1)-Rad(I);
         C:=(Two*Rad(I))*(One-Rad(I)/Rad(I+1))*(Rad(I+1)*H(I,3)-Rad(I)*H(I+1,3))/Eps+Two*Sqr(Rad(I))*Log(Rad(I+1)/Rad(I))*(H(I+1,3)-H(I,3))/Eps;
         H(I,2):=((Rad(I)/Rad(I+1))**2)*H(I+1,2)-C;
      end loop;
      H(H'First(1),2):=Comp_Zero;
      --J>=3
      for J in 3..H'Last(2)-1 loop
         JJ:=J+1;
         for I in 0..H'Last(1) loop
            H(I,J):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
         end loop;
         for I in reverse H'First(1)+1..-1 loop
            Eps:=Rad(I+1)-Rad(I);
            C:=(Two*Rad(I))/Rep(J-1)*(One-(Rad(I)/Rad(I+1))**(J-1))*(Rad(I+1)*H(I,JJ)-Rad(I)*H(I+1,JJ))/Eps+(Two*Sqr(Rad(I)))/Rep(J-2)*(One-(Rad(I)/Rad(I+1))**(J-2))*(H(I+1,JJ)-H(I,JJ))/Eps;
            H(I,J):=((Rad(I)/Rad(I+1))**J)*H(I+1,J)-C;
         end loop;
         H(H'First(1),J):=Comp_Zero;
      end loop;
      for I in H'Range(1) loop
         H(I,H'Last(2)):=Comp_Zero;--due to the second part of the  Cauchy Transfrom
      end loop;
      --J<=-2
      for J in H'First(2)..-2 loop
         H(H'First(1),J):=Comp_Zero;--value at zero
         for I in H'First(1)+1..H'Last(1) loop
            Eps:=Rad(I)-Rad(I-1);
            C:=(Two*Rad(I))/Rep(abs(J)+1)*(One-(Rad(I-1)/Rad(I))**(abs(J)+1))*(Rad(I)*H(I-1,J+1)-Rad(I-1)*H(I,J+1))/Eps+(Two*Sqr(Rad(I)))/Rep(abs(J)+2)*(One-(Rad(I-1)/Rad(I))**(abs(J)+2))*(H(I,J+1)-H(I-1,J+1))/Eps;
            if I=H'First(1)+1 then
               H(I,J):=C;
            else
               H(I,J):=((Rad(I-1)/Rad(I))**(abs(J)))*H(I-1,J)+C;
            end if;
         end loop;
      end loop;
      --J=-1
      H(H'First(1),-1):=Comp_Zero;--value at zero
      for I in H'First(1)+1..H'Last(1) loop
         Eps:=Rad(I)-Rad(I-1);
         C:=Rad(I)*(One-(Rad(I-1)/Rad(I))**2)*(Rad(I)*P(I-1)-Rad(I-1)*P(I))/Eps+(Two*Sqr(Rad(I)))/Three*(One-(Rad(I-1)/Rad(I))**3)*(P(I)-P(I-1))/Eps;
         if I=H'First(1)+1 then
            H(I,-1):=C;
         else
            H(I,-1):=(Rad(I-1)/Rad(I))*H(I-1,-1)+C;
         end if;
      end loop;
   end PP_Transform;
end Beltrami_Ops;
