with Ada.Text_IO, Reps.IO,Reps.Ops,Comps.IO, Comps.Ops,Ada.Sequential_IO,Messages;
use Ada.Text_IO,Reps.Ops,Reps.IO,Comps.IO, Comps.Ops;
use Messages;
with Lin.IO;

pragma Elaborate_All (Ada.Text_IO, Ada.Sequential_IO);

package body Lin.Ops is

   pragma Suppress(Storage_Check);

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);
   package Seq_Rep_IO is new Ada.Sequential_IO(Rep);
   use  Txt_Int_IO, Txt_Rep_IO,Seq_Rep_IO;


   procedure SetZero(A: out MatrixPtr) is
      pragma Suppress(Storage_Check);
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := Comp_Zero;
         end loop;
      end loop;
   end SetZero;


   --REAL ARRAYS
   procedure Add(B1: in RealVector; B2: in out RealVector) is
   begin
      for I in B2'Range loop
         B2(I) := B2(I)+B1(I);
      end loop;
   end Add;


   procedure Add(A1: in RealMatrix; A2: in out RealMatrix) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            A2(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
   end Add;


   procedure Add(R: in Rep; A: in out RealMatrix) is
   begin
      for I in A'Range(1) loop
         A(I,I) := A(I,I)+R;
      end loop;
   end Add;


   procedure Sub(B1: in RealVector; B2: in out RealVector) is
   begin
      for I in B2'Range loop
         B2(I) := B2(I)-B1(I);
      end loop;
   end Sub;


   procedure Sub(A1: in RealMatrix; A2: in out RealMatrix) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            A2(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
   end Sub;


   procedure Mult(R: in Rep; B: in out RealVector) is
   begin
      for I in B'Range loop
         B(I) := R*B(I);
      end loop;
   end Mult;


   procedure Mult(R: in Rep; A: in out RealMatrix) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := R*A(I,J);
         end loop;
      end loop;
   end Mult;


   procedure Mult(A1: in RealMatrix; A2: in out RealMatrix) is
      R: Rep;
      B: RealVector(A2'Range(1));
   begin
      for J in B'Range loop
         for K in B'Range loop
            B(K) := A2(K,J);
         end loop;
         for I in B'Range loop
            R := Zero;
            for K in B'Range loop
               R := R+A1(I,K)*B(K);
            end loop;
            A2(I,J) := R;
         end loop;
      end loop;
   end Mult;


   function "-"(B: RealVector) return RealVector is
      Bt: RealVector(B'Range);
   begin
      for I in B'Range loop
         Bt(I) := -B(I);
      end loop;
      return Bt;
   end "-";


   function "+"(B1,B2: RealVector) return RealVector is
      B: RealVector(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)+B1(I);
      end loop;
      return B;
   end "+";


   function "+"(A1,A2: in RealMatrix) return RealMatrix is
      A: RealMatrix(A2'First(1)..A2'Last(1),A2'First(2)..A2'Last(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
      return A;
   end "+";


   function "-"(B1,B2: RealVector) return RealVector is
      B: RealVector(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)-B1(I);
      end loop;
      return B;
   end "-";


   function "-"(A1,A2: in RealMatrix) return RealMatrix is
      A: RealMatrix(A2'First(1)..A2'Last(1),A2'First(2)..A2'Last(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
      return A;
   end "-";


   function "*"(B1,B2: RealVector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B2'Range loop
         R := R+B1(I)*B2(I);
      end loop;
      return R;
   end "*";


   function "*"(R: Rep; B: RealVector) return RealVector is
      Br: RealVector(B'Range);
   begin
      for I in B'Range loop
         Br(I) := R*B(I);
      end loop;
      return Br;
   end "*";


   function "*"(R: Rep; A: RealMatrix) return RealMatrix is
      Ar: RealMatrix(A'First(1)..A'Last(1),A'First(2)..A'Last(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Ar(I,J) := R*A(I,J);
         end loop;
      end loop;
      return Ar;
   end "*";


   function "*"(B1,B2: RealVector) return RealVector is
      B: RealVector(B2'Range);
   begin
      for I in B2'Range loop
         B(I) := B1(I)*B2(I);
      end loop;
      return B;
   end "*";


   function "*"(A: RealMatrix; B: RealVector) return RealVector is
      R: Rep;
      Br: RealVector(A'Range(1));
   begin
      for I in Br'Range loop
         R := Zero;
         for J in B'Range loop
            R := R+A(I,J)*B(J);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";


   function "*"(B: RealVector; A: RealMatrix) return RealVector is
      R: Rep;
      Br: RealVector(A'Range(2));
   begin
      for I in Br'Range loop
         R := Zero;
         for J in B'Range loop
            R := R+B(J)*A(J,I);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";


   function "*"(A1,A2: RealMatrix) return RealMatrix is
      R: Rep;
      A: RealMatrix(A1'First(1)..A1'Last(1),A1'First(2)..A1'Last(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            R := Zero;
            for K in A2'Range(1) loop
               R := R+A1(I,K)*A2(K,J);
            end loop;
            A(I,J) := R;
         end loop;
      end loop;
      return A;
   end "*";


   function Prod(B0,B1,B2: RealVector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B0'Range loop
         R := R+(B0(I)*B0(I))*B1(I)*B2(I);
      end loop;
      for I in B0'Last+1 .. B2'Last loop
         R := R+B1(I)*B2(I);
      end loop;
      return R;
   end Prod;


   function RealId(N: Positive) return RealMatrix is
      pragma Suppress(Storage_Check);
      A: RealMatrix(1..N,1..N);
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            if I=J then A(I,J):=One;
            else  A(I,J):=Zero;
            end if;
         end loop;
      end loop;
      return A;
   end RealId;


   procedure UnitVector(B: in out RealVector) is
      C: Rep;
   begin
      C := Norm2(B);
      if C>Zero then
         Mult(One/C,B);
      else
         C := One/Sqrt(Rep(B'Last-B'First+1));
         for I in B'Range loop
            B(I) := C;
         end loop;
      end if;
   end UnitVector;


   -- norms
   function Norm1(B: RealVector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+Abs(B(I));
      end loop;
      return R;
   end Norm1;


   function Norm1(A: RealMatrix) return Rep is
      R,Rj: Rep;
   begin
      R := Zero;
      for J in A'Range(1) loop
         Rj := Zero;
         for I in A'Range(2) loop
            Rj := Rj+Abs(A(I,J));
         end loop;
         if Rj>R then
            R := Rj;
         end if;
      end loop;
      return R;
   end Norm1;


   function Norm2(B: RealVector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+Abs(B(I)*B(I));
      end loop;
      return Sqrt(R);
   end Norm2;


   --regular solvers
   procedure Decomp(A: in out RealMatrix; Cond: out Rep; Pvt: out Pivot; Work: out RealVector) is
      -- translated from numerical recipes
      N0,N,M: Integer;
      T,Anorm,Ynorm,Znorm: Rep;
   begin
      N0 := Work'First;
      N := Work'Last;
      Pvt(N) := N0;
      if N = N0 then
         Cond := One;
         if A(N0,N0) = Zero then
            Message("Lin.Ops.Decomp1",Matrix_Singular);
            return;
         end if;
         return;
      end if;

      Anorm := Zero;
      for J in N0 .. N loop
         T := Zero;
         for I in N0 .. N loop
            T := T+Abs(A(I,J));
         end loop;
         if T > Anorm then
            Anorm := T;
         end if;
      end loop;

      for K in N0 .. N-1 loop
         M := K;
         for I in K+1 .. N loop
            if Abs(A(I,K)) > Abs(A(M,K)) then
               M := I;
            end if;
         end loop;
         Pvt(K) := M;
         if M /= K then
            Pvt(N) := -Pvt(N);
         end if;

         T := A(M,K);
         A(M,K) := A(K,K);
         A(K,K) := T;
         if T /= Zero then
            for I in K+1 .. N loop
               A(I,K) := -A(I,K)/T;
            end loop;
            for J in K+1 .. N loop
               T := A(M,J);
               A(M,J) := A(K,J);
               A(K,J) := T;
               if T /= Zero then
                  for I in K+1 .. N loop
                     A(I,J) := A(I,J)+A(I,K)*T;
                  end loop;
               end if;
            end loop;
         end if;
      end loop;

      for K in N0 .. N loop
         T := Zero;
         for I in N0 .. K-1 loop
            T := T+A(I,K)*Work(I);
         end loop;
         if A(K,K) = Zero then
            Message("Lin.Ops.Decomp2",Matrix_Singular);
            return;
         end if;
         if T >= Zero then
            Work(K) := -(T+One)/A(K,K);
         else
            Work(K) := -(T-One)/A(K,K);
         end if;
      end loop;

      for K in reverse N0 .. N-1 loop
         T := Zero;
         for I in K+1 .. N loop
            T := T+A(I,K)*Work(K);
         end loop;
         Work(K) := T;
         M := Pvt(K);
         if M /= K then
            T := Work(M);
            Work(M) := Work(K);
            Work(K) := T;
         end if;
      end loop;

      Ynorm := Zero;
      for I in N0 .. N loop
         Ynorm := Ynorm+Abs(Work(I));
      end loop;

      Solve(A,Work,Pvt);

      Znorm := Zero;
      for I in N0 .. N loop
         Znorm := Znorm+Abs(Work(I));
      end loop;

      Cond := Anorm*Znorm/Ynorm;
      if Cond < One then
         Cond := One;
      end if;
   end Decomp;


   procedure Solve(A: in RealMatrix; B: in out RealVector; Pvt: in Pivot) is
      -- solve A*B2=B1
      -- input:   A     triangularized matrix obtained from Decomp
      --          Pvt   pivot vector obtained from Decomp
      --          B     vector B1
      -- output:  B     vector B2
      N0,N,M: Integer;
      T: Rep;
   begin
      N0 := B'First;
      N := B'Last;

      for K in N0 .. N-1 loop
         M := Pvt(K);
         T := B(M);
         B(M) := B(K);
         B(K) := T;
         for I in K+1 .. N loop
            B(I) := B(I)+A(I,K)*T;
         end loop;
      end loop;

      for K in reverse N0+1 .. N loop
         B(K) := B(K)/A(K,K);
         T := -B(K);
         for I in N0 .. K-1 loop
            B(I) := B(I)+A(I,K)*T;
         end loop;
      end loop;

      B(N0) := B(N0)/A(N0,N0);
   end Solve;


   procedure Solve(W: in RealMatrix; B: in RealVector; X: in out RealVector; Rsq: out Rep; Eps: in Rep := 1.0E-5) is
      -- translated procedure Sparse from numerical recipes
      Vsize: Integer;
      Rp,Gg,Gam,Eps2,Dgg,Bsq,Anum,Aden: Rep;
      G,H,Xi,Xj: RealVector(B'Range);
   begin
      Vsize := B'Last-B'First+1;
      Eps2 := Rep(Vsize)*Eps*Eps;
      for Irst in 1 .. 30 loop    -- restart at most 3 times
         Xi := W*X;
         Rp := Zero;
         Bsq := Zero;
         for J in B'Range loop
            Bsq := Bsq+B(J)*B(J);
            Xi(J) := Xi(J)-B(J);
            Rp := Rp+Xi(J)*Xi(J);
         end loop;
         G := Xi*W;
         for J in B'Range loop
            G(J) := -G(J);
            H(J) := G(J);
         end loop;
         for Chances in reverse 1 .. 100000*Vsize loop
            Xi := W*H;
            Anum := Zero;
            Aden := Zero;
            for J in B'Range loop
               Anum := Anum+G(J)*H(J);
               Aden := Aden+Xi(J)*Xi(J);
            end loop;
            if Aden = Zero then
               Message("Lin.Ops.Solve",Matrix_Singular);
               return;
            end if;
            Anum := Anum/Aden;
            for J in B'Range loop
               Xi(J) := X(J);
               X(J) := X(J)+Anum*H(J);
            end loop;
            Xj := W*X;
            Rsq := Zero;
            for J in B'Range loop
               Xj(J) := Xj(J)-B(J);
               Rsq := Rsq+Xj(J)*Xj(J);
            end loop;
            if Rsq=Rp or Rsq<=Bsq*Eps2 then return; end if;
            if Rsq>Rp then
               X := Xi;
               exit when True;               --- restart
            end if;
            Rp := Rsq;
            Xi := Xj*W;
            Gg := Zero;
            Dgg := Zero;
            for J in B'Range loop
               Gg := Gg+G(J)*G(J);
               Dgg := Dgg+(Xi(J)+G(J))*Xi(J);
            end loop;
            if Gg=Zero then return; end if;
            Gam := Dgg/Gg;
            for J in B'Range loop
               G(J) := -Xi(J);
               H(J) := G(J)+Gam*H(J);
            end loop;
            if Chances=1 then
               Put("Too many iterations in Solve");
               raise Getting_Tired;
            end if;
         end loop;
      end loop;
   end Solve;


   procedure Invert(A: in out RealMatrix; Output: in Boolean:=False) is
      C: Rep;
      Pvt: Pivot(A'Range(1));
      B,BB: RealVector(A'Range(1));
      Aux,AA: RealMatrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      AA:=A;
      Decomp(Aux,C,Pvt,B);
      for J in A'Range(2) loop
         SetZero(B);
         B(J) := One;
         BB:=B;
         Solve(Aux,B,Pvt);
         if Output then
            New_Line;
            Put("l_1 norm in Invert.Lin-Ops: ");Txt_Put(Current_Output,Norm1(AA*B-BB));
         end if;
         SetColumn(J,B,A);
      end loop;
   end Invert;


   procedure ImproveInv(A: in  RealMatrix; InvA: in out RealMatrix; Err: in Rep := 1.0E-5; Output: in Boolean:=False) is
      Iter: constant Integer := 20000000;
      Er: Rep;
      Iden: RealMatrix(A'Range(1),A'Range(2));
   begin
      if A'Length(1)/=A'Length(2)  then
         Message("ImproveInv.Lin-Ops",Index_Error);
      end if;

      SetZero(Iden);
      for I in Iden'Range(1) loop
         Iden(I,I):=One;
      end loop;

      for I in 1 .. Iter loop
         InvA:=InvA*A*InvA-Two*InvA;
         Er:=Norm1(A*InvA-Iden);
         if Output then
            New_Line;
            Put("l_1 norm in ImproveInv.Lin-Ops: ");Txt_Put(Current_Output,Er);
         end if;
         if Er<Err then
            New_Line;
            Put("l_1 norm in ImproveInv.Lin-Ops: ");Txt_Put(Current_Output,Er);
         end if;
      end loop;
      Put(Current_Output,Iter,3,10);
      Put(" iterations in Eigenvector");
      raise Getting_Tired;
   end ImproveInv;


   --eigensolvers
   procedure Balanc(A: in out RealMatrix) is
      -- translated from numerical recipes
      NearOne: constant Rep := 0.95;
      Radix: constant Rep := Two;
      Done: Boolean;
      N0,N: Integer;
      S,R,G,F,C,Sqrdx: Rep;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      Sqrdx := Radix*Radix;
      loop
         Done := True;
         for I in N0 .. N loop
            C := Zero;
            R := Zero;
            for J in N0 .. N loop
               if J /= I then
                  C := C+Abs(A(J,I));
                  R := R+Abs(A(I,J));
               end if;
            end loop;
            if C /= Zero and then R /= Zero then
               G := R/Radix;
               F := One;
               S := C+R;
               while C < G loop
                  F := F*Radix;
                  C := C*Sqrdx;
               end loop;
               G := R*Radix;
               while C > G loop
                  F := F/Radix;
                  C := C/Sqrdx;
               end loop;
               if (C+R)/F < NearOne*S then
                  Done := False;
                  G := One/F;
                  for J in N0 .. N loop A(I,J) := A(I,J)*G; end loop;
                  for J in N0 .. N loop A(J,I) := A(J,I)*F; end loop;
               end if;
            end if;
         end loop;
         exit when Done;
      end loop;
   end Balanc;


   procedure ElmHes(A: in out RealMatrix) is
      -- ugly, translated from numerical recipes
      N0,N,L: Integer;
      X,Y: Rep;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      if N-N0<2 then return; end if;
      for M in N0+1 .. N-1 loop
         X := Zero;
         L := M;
         for J in M .. N loop
            if Abs(A(J,M-1)) > Abs(X) then
               X := A(J,M-1);
               L := J;
            end if;
         end loop;
         if L /= M then
            for J in M-1 .. N loop
               Y := A(L,J);
               A(L,J) := A(M,J);
               A(M,J) := Y;
            end loop;
            for J in N0 .. N loop
               Y := A(J,L);
               A(J,L) := A(J,M);
               A(J,M) := Y;
            end loop;
         end if;
         if X /= Zero then
            for I in M+1 .. N loop
               Y := A(I,M-1);
               if Y /= Zero then
                  Y := Y/X;
                  A(I,M-1) := Y;
                  for J in M .. N loop A(I,J) := A(I,J)-Y*A(M,J); end loop;
                  for J in N0 .. N loop A(J,M) := A(J,M)+Y*A(J,I); end loop;
               end if;
            end loop;
         end if;
      end loop;
   end ElmHes;


   procedure Hqr(A: in out RealMatrix; Wr,Wi: out RealVector) is
      -- ugly, translated from numerical recipes
      function Sign(R,S: Rep) return Rep is
      begin
         if S<Zero then
            return -Abs(R);
         else
            return Abs(R);
         end if;
      end Sign;

      C1: constant Rep := 0.75;
      C2: constant Rep := -0.4375;
      FoundRoot: Boolean;
      N0,N,Nn,M,L,Its: Integer;
      Z,Y,X,W,V,U,T,S,R,Q,P,Anorm: Rep;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      Anorm := Abs(A(N0,N0));
      for I in N0+1 .. N loop
         for J in I-1 .. N loop
            Anorm := Anorm+Abs(A(I,J));
         end loop;
      end loop;
      Nn := N;
      T := Zero;
      while Nn>=N0 loop
         FoundRoot := False;
         Its := 0;
         while not FoundRoot loop
            L := Nn;
            while L>N0 loop
               S := Abs(A(L-1,L-1))+Abs(A(L,L));
               if S=Zero then
                  S := Anorm;
               end if;
               exit when (Abs(A(L,L-1))+S) = S;
               L := L-1;
            end loop;
            X := A(Nn,Nn);
            if L=Nn then
               FoundRoot := True;
               Wr(Nn) := X+T;
               Wi(Nn) := Zero;
               Nn := Nn-1;
            else
               Y := A(Nn-1,Nn-1);
               W := A(Nn,Nn-1)*A(Nn-1,Nn);
               if L = Nn-1 then
                  FoundRoot := True;
                  P := Half*(Y-X);
                  Q := P*P+W;
                  Z := Sqrt(Abs(Q));
                  X := X+T;
                  if Q >= Zero then
                     Z := P+Sign(Z,P);
                     Wr(Nn) := X+Z;
                     Wr(Nn-1) := Wr(Nn);
                     if Z /= Zero then Wr(Nn) := X-W/Z; end if;
                     Wi(Nn) := Zero;
                     Wi(Nn-1) := Zero;
                  else
                     Wr(Nn) := X+P;
                     Wr(Nn-1) := Wr(Nn);
                     Wi(Nn) := Z;
                     Wi(Nn-1) := -Z;
                  end if;
                  Nn := Nn-2;
               else
                  if Its=10 or Its=20 then
                     T := T+X;
                     for I in N0 .. Nn loop
                        A(I,I) := A(I,I)-X;
                     end loop;
                     S := Abs(A(Nn,Nn-1))+Abs(A(Nn-1,Nn-2));
                     X := C1*S;
                     Y := X;
                     W := C2*S*S;
                  elsif Its = 30 then
                     Put("Too many iterations in Hqr");
                     raise Getting_Tired;
                  end if;
                  Its := Its+1;
                  M := Nn-2;
                  if M>=L then
                     loop
                        Z := A(M,M);
                        R := X-Z;
                        S := Y-Z;
                        P := (R*S-W)/A(M+1,M)+A(M,M+1);
                        Q := A(M+1,M+1)-Z-R-S;
                        R := A(M+2,M+1);
                        S := Abs(P)+Abs(Q)+Abs(R);
                        P := P/S;
                        Q := Q/S;
                        R := R/S;
                        exit when M=L;
                        U := Abs(A(M,M-1))*(Abs(Q)+Abs(R));
                        V := Abs(P)*(Abs(A(M-1,M-1))+Abs(Z)+Abs(A(M+1,M+1)));
                        exit when U+V=V;
                        M := M-1;
                     end loop;
                  end if;
                  for I in M+2 .. Nn loop
                     A(I,I-2) := Zero;
                     if  I /= (M+2) then A(I,I-3) := Zero; end if;
                  end loop;
                  for K in M .. Nn-1 loop
                     if K /= M then
                        P := A(K,K-1);
                        Q := A(K+1,K-1);
                        R := Zero;
                        if K /= (Nn-1) then
                           R := A(K+2,K-1);
                        end if;
                        X := Abs(P)+Abs(Q)+Abs(R);
                        if X /= Zero then
                           P := P/X;
                           Q := Q/X;
                           R := R/X;
                        end if;
                     end if;
                     S := Sign(Sqrt(P*P+Q*Q+R*R),P);
                     if S /= Zero then
                        if K=M then
                           if L /= M then A(K,K-1) := -A(K,K-1); end if;
                        else
                           A(K,K-1) := -S*X;
                        end if;
                        P := P+S;
                        X := P/S;
                        Y := Q/S;
                        Z := R/S;
                        Q := Q/P;
                        R := R/P;
                        for J in K .. Nn loop
                           P := A(K,J)+Q*A(K+1,J);
                           if K /= (Nn-1) then
                              P := P+R*A(K+2,J);
                              A(K+2,J) := A(K+2,J)-P*Z;
                           end if;
                           A(K+1,J) := A(K+1,J)-P*Y;
                           A(K,J) := A(K,J)-P*X;
                        end loop;
                        for I in L .. Integer'Min(Nn,K+3) loop
                           P := X*A(I,K)+Y*A(I,K+1);
                           if K /= (Nn-1) then
                              P := P+Z*A(I,K+2);
                              A(I,K+2) := A(I,K+2)-P*R;
                           end if;
                           A(I,K+1) := A(I,K+1)-P*Q;
                           A(I,K) := A(I,K)-P;
                        end loop;
                     end if;
                  end loop;
               end if;
            end if;
         end loop;
      end loop;
   end Hqr;


   procedure Eigenvector(A: in RealMatrix; R: in Rep; B: in out RealVector; Eps: in Rep := 1.0E-8) is
      Iter: constant Integer := 200000;
      C: Rep;
      Pvt: Pivot(B'Range);
      Bt: RealVector(B'Range);
      Aux: RealMatrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      Add(-R,Aux);
      Decomp(Aux,C,Pvt,Bt);
      UnitVector(B);
      for I in 1 .. Iter loop
         Bt := B;
         Solve(Aux,B,Pvt);
         C := One/Norm2(B);
         if B*Bt<Zero then C := -C; end if;
         Mult(C,B);
         Sub(B,Bt);
         if Norm2(Bt)<Abs(Eps) then return; end if;
      end loop;
      Txt_Int_IO.Put(Current_Output,Iter,4,10);
      Put(" iterations in EigenVector");
      if Eps<Zero then return; end if;               --- for hack
      raise Getting_Tired;
   end Eigenvector;


   procedure Eigenvalues(A: in RealMatrix; Br,Bi: out RealVector) is
      Aux: RealMatrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      Balanc(Aux);
      ElmHes(Aux);
      Hqr(Aux,Br,Bi);
   end Eigenvalues;


   --descent and conjugate methods
   procedure Descent(W: in RealMatrix; B: in RealVector; X: in out RealVector; Eps: in Rep := 1.0E-8) is
      -- a bit primitive
      Vsize: Integer;
      Err,Lambda: Rep;
      R: RealVector(B'Range);
   begin
      Vsize := B'Last-B'First+1;
      for I in 1 .. 100*Vsize loop
         R := (W*X-B)*W;
         Err := Norm2(R);
         Lambda := Sqr(Err/Norm2(W*R));
         X := X-Lambda*R;
         New_Line;
         Txt_Put(Current_Output, Abs(Lambda)*Err);
         if  Abs(Lambda)*Err<Eps then return; end if;
      end loop;
      Put_Line("Linear_Ops.Descent: Giving up");
   end Descent;


   procedure BCG(W: in RealMatrix; B: in RealVector; X: in out RealVector; Eps: in Rep := 1.0E-8) is
      Bf: RealVector(1 .. 1);
   begin
      Bf(1) := One;
      BCG(W,Bf,B,X,Eps,Eps);
   end BCG;

   procedure BCG(W: in RealMatrix; Bf,B: in RealVector; X: in out RealVector; Eabs,Erel: in Rep) is
      I,Vsize: Integer;
      Rho,Rhold,Alpha,Beta,Eps,Err: Rep;
      R,Rt,P,Pt,Q,Qt: RealVector(B'Range);
   begin
      Put("Linear.Ops.BCG: ");
      Vsize := B'Last-B'First+1;
      R := B-W*X;
      Err := Sqrt(Prod(Bf,R,R));
      Eps := Rep'Min(Eabs,Erel*Err);
      Rt := R;
      I := 1;
      Rho := Zero;
      Alpha := Zero;
      loop
         Rhold := Rho;
         Rho := Rt*R;
         if I=1 then
            P := R;
            Pt := Rt;
         else
            Beta := Rho/Rhold;
            for J in B'Range loop
               P(J) := Beta*P(J)+R(J);
               Pt(J) := Beta*Pt(J)+Rt(J);
            end loop;
         end if;
         Q := W*P;
         Qt := Pt*W;
         Alpha := Rho/(Pt*Q);
         for J in B'Range loop
            X(J) := X(J)-Alpha*P(J);
            R(J) := R(J)-Alpha*Q(J);
            Rt(J) := Rt(J)-Alpha*Qt(J);
         end loop;
         Err := Sqrt(Prod(Bf,R,R));
         exit when I=200*Vsize or Err<Eps;
         I := I+1;
      end loop;
      Txt_Int_IO.Put(Current_Output,I,4,10);
      Put(" iterations, Err =");
      Txt_Rep_IO.Put(Current_Output,Err,2,5,4);        -- File,Item,Fore,Aft,Exp
      New_Line;
   end BCG;


   procedure BCGstab(W: in RealMatrix; B: in RealVector; X: in out RealVector; Eps: in Rep := 1.0E-8) is
      Bf: RealVector(1 .. 1);
   begin
      Bf(1) := One;
      BCGstab(W,Bf,B,X,Eps,Eps);
   end BCGstab;


   procedure BCGstab(W: in RealMatrix; Bf,B: in RealVector; X: in out RealVector; Eabs,Erel: in Rep) is
      I,Vsize: Integer;
      Eps,Rho,Rhold,Alpha,Beta,Gamma,Err: Rep;
      R,Rt,P,Q,T: RealVector(B'Range);
   begin
      Put("Linear.Ops.BCGstab: ");
      Vsize := B'Last-B'First+1;
      R := B-W*X;
      Rt := R;
      Err := Sqrt(Prod(Bf,R,R));
      Eps := Rep'Min(Eabs,Erel*Err);
      I := 1;
      Rho := Zero;
      Alpha := Zero;
      Gamma := Zero;
      Q(1) := Zero;    -- just to avoid compiler warnings
      loop
         Rhold := Rho;
         Rho := Prod(Bf,Rt,R);
         if I=1 then
            P := R;
         else
            Beta := (Rho/Rhold)*(Alpha/Gamma);
            for J in B'Range loop
               P(J) := R(J)+Beta*(P(J)-Gamma*Q(J));
            end loop;
         end if;
         Q := W*P;
         Alpha := Rho/Prod(Bf,Rt,Q);
         for J in B'Range loop
            R(J) := R(J)-Alpha*Q(J);
         end loop;
         T := W*R;
         Gamma := Prod(Bf,T,R)/Prod(Bf,T,T);
         for J in B'Range loop
            X(J) := X(J)-Alpha*P(J)-Gamma*R(J);
            R(J) := R(J)-Gamma*T(J);
         end loop;
         Err := Sqrt(Prod(Bf,R,R));
         exit when I=200*Vsize or Err<Eps;
         I := I+1;
      end loop;
      Txt_Int_IO.Put(Current_Output,I,4,10);
      Put(" iterations, Err =");
      Txt_Rep_IO.Put(Current_Output,Err,2,5,4);        -- File,Item,Fore,Aft,Exp
      New_Line;
   end BCGstab;


   --singular value decomposition
   -- This algorithm is based on the svd() routine in EISPACK, available at
   -- http://www.netlib.org/eispack/svd.f
   procedure SVD( A: in RealMatrix;U: out RealMatrix; W: out RealMatrix; V: out RealMatrix) is
      function MultSign(a, b: in Rep) return Rep is
      begin
         if b >= 0.0 then
            return abs a;
         else
            return -(abs a);
         end if;
      end MultSign;
      function Min(a, b: in Integer) return Integer is
      begin
         if a > b then return b; else return a; end if;
      end Min;
      function Max(a, b: in Rep) return Rep is
      begin
         if a > b then return a; else return b; end if;
      end Max;
      function Pythag(a, b: in Rep) return Rep is
         absa, absb: Rep;
      begin
         absa := abs a;
         absb := abs b;
         if absa > absb then
            return absa*Sqrt(One+(absb/absa)**2);
         elsif absb /= 0.0 then
            return absb*Sqrt(One+(absa/absb)**2);
         else
            return Zero;
         end if;
      end Pythag;

      umat: RealMatrix(1..A'Length(1), 1..A'Length(2));
      wmat: RealMatrix(umat'Range(2), umat'Range(2)) := (others => (others => 0.0));
      vmat: RealMatrix(umat'Range(2), umat'Range(2));
      rv1: RealVector(umat'Range(2));
      Flag: Boolean;
      l,nm: Integer;
      anorm,c,f,g,h,s,scale,x,y,z: Rep;

      m: Integer := umat'Last(1);
      n: Integer := umat'Last(2);
   begin
      if A'Length(1) /= A'Length(2) then
         Message("Lin.Ops.SVD",Index_Error);
         return;
      end if;

      umat := A;

      g := Zero;
      scale := Zero;
      anorm := Zero;
      for I in 1..n loop
         l := I+1;
         rv1(I) := scale*g;
         g := Zero;
         s := Zero;
         scale := Zero;
         if (I <= m) then
            for K in I..m loop
               scale := (scale + abs umat(K,I));
            end loop;
            if (scale /= Zero) then
               for K in I..m loop
                  umat(K,I) := (umat(K,I) / scale);
                  s := (s + umat(K,I)*umat(K,I));
               end loop;
               f := umat(I,I);
               g := -MultSign(Sqrt(s),f);
               h := f*g - s;
               umat(I,I) := f-g;
               for J in l..n loop
                  s := Zero;
                  for K in I..m loop
                     s := (s + umat(K,I)*umat(K,J));
                  end loop;
                  f := s/h;
                  for k in i..m loop
                     umat(k,j) := (umat(k,j) + f*umat(k,i));
                  end loop;
               end loop;
               for k in i..m loop
                  umat(k,i) := (umat(k,i) * scale);
               end loop;
            end if;
         end if;
         wmat(i,i) := (scale * g);
         g := Zero;
         s := Zero;
         scale := Zero;
         if (i <= m) and (i /= n) then
            for k in l..n loop
               scale := (scale + abs umat(i,k));
            end loop;
            if (scale /= Zero) then
               for k in l..n loop
                  umat(i,k) := (umat(i,k) / scale);
                  s := (s + umat(i,k)*umat(i,k));
               end loop;
               f := umat(i,l);
               g := -MultSign(Sqrt(s),f);
               h := f*g-s;
               umat(i,l) := f-g;
               for k in l..n loop
                  rv1(k) := umat(i,k)/h;
               end loop;
               for j in l..m loop
                  s := Zero;
                  for k in l..n loop
                     s := (s + umat(j,k)*umat(i,k));
                  end loop;
                  for k in l..n loop
                     umat(j,k) := (umat(j,k) + s*rv1(k));
                  end loop;
               end loop;
               for k in l..n loop
                  umat(i,k) := (umat(i,k) * scale);
               end loop;
            end if;
         end if;
         anorm := Max(anorm, abs wmat(i,i) + abs rv1(i));
      end loop;
      for i in reverse 1..n loop
         if (i < n) then
            if (g /= Zero) then
               for j in l..n loop
                  vmat(j,i) := (umat(i,j) / umat(i,l)) / g;
               end loop;
               for j in l..n loop
                  s := Zero;
                  for k in l..n loop
                     s := (s + umat(i,k)*vmat(k,j));
                  end loop;
                  for k in l..n loop
                     vmat(k,j) := (vmat(k,j) + s*vmat(k,i));
                  end loop;
               end loop;
            end if;
            for j in l..n loop
               vmat(i,j) := Zero;
               vmat(j,i) := Zero;
            end loop;
         end if;
         vmat(i,i) := One;
         g := rv1(i);
         l := i;
      end loop;
      for i in reverse 1..Min(m,n) loop
         l := i+1;
         g := wmat(i,i);
         for j in l..n loop
            umat(i,j) := Zero;
         end loop;
         if (g /= Zero) then
            g := One/g;
            for j in l..n loop
               s := Zero;
               for k in l..m loop
                  s := (s + umat(k,i)*umat(k,j));
               end loop;
               f := (s/umat(i,i))*g;
               for k in i..m loop
                  umat(k,j) := (umat(k,j) + f*umat(k,i));
               end loop;
            end loop;
            for j in i..m loop
               umat(j,i) := (umat(j,i)*g);
            end loop;
         else
            for j in i..m loop
               umat(j,i) := Zero;
            end loop;
         end if;
         umat(i,i) := umat(i,i) + One;
      end loop;
      for k in reverse 1..n loop
         for its in 1..30 loop
            Flag := True;
            for ll in reverse 1..k loop
               l := ll;
               nm := l-1;
               if (abs rv1(l) + anorm) = anorm then
                  Flag := False;
                  exit;
               elsif (abs wmat(nm,nm) + anorm) = anorm then
                  exit;
               end if;
            end loop;
            if Flag then
               c := Zero;
               s := One;
               for i in l..k loop
                  f := s*rv1(i);
                  rv1(i) := c*rv1(i);
                  if (abs f + anorm) = anorm then
                     exit;
                  end if;
                  g := wmat(i,i);
                  h := pythag(f,g);
                  wmat(i,i) := h;
                  h := One/h;
                  c := g*h;
                  s := -f*h;
                  for j in 1..m loop
                     y := umat(j,nm);
                     z := umat(j,i);
                     umat(j,nm) := y*c+z*s;
                     umat(j,i) := z*c-y*s;
                  end loop;
               end loop;
            end if;
            z := wmat(k,k);
            if l = k then
               if z < Zero then
                  wmat(k,k) := -z;
                  for j in 1..n loop
                     vmat(j,k) := -vmat(j,k);
                  end loop;
               end if;
               exit;
            end if;
            if (its = 30) then
               Put("Too many iterations in Hqr");
               raise Getting_Tired;
            end if;
            x := wmat(l,l);
            nm := k-1;
            y := wmat(nm,nm);
            g := rv1(nm);
            h := rv1(k);
            f := ((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
            g := pythag(f,One);
            f := ((x-z)*(x+z)+h*((y/(f+MultSign(g,f)))-h))/x;
            c := One;
            s := One;
            for j in l..nm loop
               g := rv1(j+1);
               y := wmat(j+1,j+1);
               h := s*g;
               g := c*g;
               z := pythag(f,h);
               rv1(j) := z;
               c := f/z;
               s := h/z;
               f := x*c+g*s;
               g := g*c-x*s;
               h := y*s;
               y := y*c;
               for jj in 1..n loop
                  x := vmat(jj,j);
                  z := vmat(jj,j+1);
                  vmat(jj,j) := x*c+z*s;
                  vmat(jj,j+1) := z*c-x*s;
               end loop;
               z := pythag(f,h);
               wmat(j,j) := z;
               if (z /= Zero) then
                  z := One/z;
                  c := f*z;
                  s := h*z;
               end if;
               f := c*g+s*y;
               x := c*y-s*g;
               for jj in 1..m loop
                  y := umat(jj,j);
                  z := umat(jj,j+1);
                  umat(jj,j) := y*c+z*s;
                  umat(jj,j+1) := z*c-y*s;
               end loop;
            end loop;
            rv1(l) := Zero;
            rv1(k) := f;
            wmat(k,k) := x;
         end loop;
      end loop;
      u := umat;
      w := wmat;
      v := vmat;
   end SVD;


   procedure SVDSolve(A: in  RealMatrix; X: in out RealVector; Output: in Boolean:=False) is
      pragma Suppress(Storage_Check);
      U,W,V: RealMatrix(A'Range(1),A'Range(2));
      B: RealVector(X'Range);
   begin
      B:=X;

      if A'First(2)/=X'First(1) or  A'Last(2)/=X'Last(1)  then
         Message("SVDSolve.Lin-Ops",Index_Error);
      end if;

      SVD(A,U,W,V);
      Transpose(U);
      X:=U*X;
      for I in X'Range loop
         X(I):=X(I)/W(I,I);
      end loop;
      X:=V*X;
      if Output then
         New_Line;
         Put("l_1 norm in SVDSolve.Lin-Ops: ");Txt_Put(Current_Output,Norm1(A*X-B));
      end if;
   end SVDSolve;


   procedure InvertSVD(A: in out RealMatrix; Output: in Boolean:=False) is
      pragma Suppress(Storage_Check);
      B,BB: RealVector(A'Range(1));
      AA,W,U,V: RealMatrix(A'Range(1),A'Range(2));
   begin
      AA:=A;
      SVD(A,U,W,V);
      for J in A'Range(2) loop
         SetZero(BB);
         BB(J):=One;
         for I in A'Range(2) loop
            B(I):=U(J,I)/W(I,I);
         end loop;
         B:=V*B;
         if Output then
            New_Line;
            Put("l_1 norm in InvertSVD.Lin-Ops: ");Txt_Put(Current_Output,Norm1(AA*B-BB));
         end if;
         SetColumn(J,B,A);
      end loop;
   end InvertSVD;


   --COMPLEX ARRAYS
   procedure Add(B1: in Vector; B2: in out Vector) is
   begin
      for I in B2'Range loop
         B2(I) := B2(I)+B1(I);
      end loop;
   end Add;


   procedure Add(A1: in Matrix; A2: in out Matrix) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            A2(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
   end Add;


   procedure Add(A1: in MatrixPtr; A2: in out MatrixPtr) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            A2(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
   end Add;


   function Norm2(B: Vector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+AbsVal(B(I)*B(I));
      end loop;
      return Sqrt(R);
   end Norm2;


   procedure Add(R: in Comp; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
         A(I,I) := A(I,I)+R;
      end loop;
   end Add;


   procedure Sub(B1: in Vector; B2: in out Vector) is
   begin
      for I in B2'Range loop
         B2(I) := B2(I)-B1(I);
      end loop;
   end Sub;


   procedure Sub(A1: in Matrix; A2: in out Matrix) is
   begin
      for I in A2'Range(1) loop
         for J in A2'Range(2) loop
            A2(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
   end Sub;


   procedure Mult(R: in Comp; B: in out Vector) is
   begin
      for I in B'Range loop
         B(I) := R*B(I);
      end loop;
   end Mult;


   procedure Mult(R: in Comp; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := R*A(I,J);
         end loop;
      end loop;
   end Mult;


   procedure Mult(A1: in Matrix; A2: in out Matrix) is
      R: Comp;
      B: Vector(A2'Range(1));
   begin
      for J in B'Range loop
         for K in B'Range loop
            B(K) := A2(K,J);
         end loop;
         for I in B'Range loop
            R := Comp_Zero;
            for K in B'Range loop
               R := R+A1(I,K)*B(K);
            end loop;
            A2(I,J) := R;
         end loop;
      end loop;
   end Mult;


   function Less(L,R,Base: in Comp) return Boolean is
   begin
      if AbsVal(L-Base)< AbsVal(R-Base) then
         return True;
      else
         return False;
      end if;
   end Less;


   procedure Swap(X,Y: in out Comp) is
      T : Comp;
   begin
      T := X;
      X := Y;
      Y := T;
   end Swap;


   procedure Print_Marker(M : String; P : Integer) is
   begin
      for I in 1 .. P-1 loop
         Put("    ");
      end loop;
      Put("   " & M);
      New_Line;
   end Print_Marker;


   -- This procedure will sift down the From'th element of the array assuming
   -- that the rest of the array already is a heap.
   procedure Sift_down(A : in out Vector;
                       From : Positive; Base: in Comp) is
      I,N,L,H : Integer;
   begin
      pragma Assert(From >= A'First and From <= A'Last);

      I := From;

      -- While the index is parent to any object in the array.
      while 2 * (I - A'First + 1) + A'First - 1 <= A'Last loop

         -- Compute index of low child
         L := 2 * (I - A'First + 1) + A'First - 1;
         -- Compute index of high child
         H := 2 * (I - A'First + 1) + A'First;

         N := I;
         if Less(A(N),A(L),Base) then
            N := L;
         end if;
         if H <= A'Last and then Less(A(N),A(H),Base) then
            N := H;
         end if;
         exit when N = I;
         Swap(A(I), A(N));
         I := N;
      end loop;
   end Sift_down;


   -- This procedure will sift up the last element of the array assuming
   -- that the rest of the array already is a heap.
   procedure Sift_Up(A : in out Vector; Base: in Comp) is
      I,P : Integer;
   begin
      I := A'Last;

      -- While the index is parent to any object in the array.
      while I > A'First loop
         P := (I - A'First + 1) / 2 + A'First - 1;

         exit when not Less(A(I),A(P),Base);    -- To use "<" only

         Swap(A(I), A(P));
         I := P;
      end loop;
   end Sift_up;


   procedure Heap_Sort(A : in out Vector; Base: in Comp) is
   begin
      -- Heapifying array
      for I in reverse A'First .. (A'Length / 2 + A'First - 1) loop
         --      Print_Array(A);
         --      Print_Marker("^",I);
         Sift_Down(A,I,Base);
      end loop;
      --        New_Line; Print_Array(A); New_Line;
      -- Sorting array
      for I in reverse A'First .. A'Last loop
         Swap(A(A'First),A(I));
         Sift_Down(A(A'First .. (I-1)), A'First,Base);
         --      Print_Array(A);
         --      Print_Marker("|",I);
      end loop;
   end Heap_Sort;


   function "-"(B: Vector) return Vector is
      Bt: Vector(B'Range);
   begin
      for I in B'Range loop
         Bt(I) := Neg(B(I));
      end loop;
      return Bt;
   end "-";


   function "+"(B1,B2: Vector) return Vector is
      B: Vector(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)+B1(I);
      end loop;
      return B;
   end "+";


   function "+"(A1,A2: in Matrix) return Matrix is
      A: Matrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
      return A;
   end "+";


   function "-"(B1,B2: Vector) return Vector is
      B: Vector(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)-B1(I);
      end loop;
      return B;
   end "-";


   function "-"(A1,A2: in Matrix) return Matrix is
      A: Matrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
      return A;
   end "-";


   function "*"(B1,B2: Vector) return Comp is
      R: Comp;
   begin
      R := Comp_Zero;
      for I in B2'Range loop
         R := R+B1(I)*Conj(B2(I));
      end loop;
      return R;
   end "*";


   function "*"(R: Comp; B: Vector) return Vector is
      Br: Vector(B'Range);
   begin
      for I in B'Range loop
         Br(I) := R*B(I);
      end loop;
      return Br;
   end "*";


   function "*"(R: Comp; A: Matrix) return Matrix is
      Ar: Matrix(A'Range(1),A'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Ar(I,J) := R*A(I,J);
         end loop;
      end loop;
      return Ar;
   end "*";


   function "*"(B1,B2: Vector) return Vector is
      B: Vector(B2'Range);
   begin
      for I in B2'Range loop
         B(I) := B1(I)*Conj(B2(I));
      end loop;
      return B;
   end "*";


   function "*"(A: Matrix; B: Vector) return Vector is
      R: Comp;
      Br: Vector(A'Range(1));
   begin
      for I in Br'Range loop
         R := Comp_Zero;
         for J in B'Range loop
            R := R+A(I,J)*B(J);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";


   function "*"(B: Vector; A: Matrix) return Vector is
      R: Comp;
      Br: Vector(A'Range(2));
   begin
      for I in Br'Range loop
         R := Comp_Zero;
         for J in B'Range loop
            R := R+B(J)*A(J,I);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";


   function "*"(A1,A2: Matrix) return Matrix is
      R: Comp;
      A: Matrix(A1'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            R := Comp_Zero;
            for K in A2'Range(1) loop
               R := R+A1(I,K)*A2(K,J);
            end loop;
            A(I,J) := R;
         end loop;
      end loop;
      return A;
   end "*";


   function "*"(W: Sparse; B: Vector) return Vector is
      Br: Vector(B'Range);
   begin
      SetZero(Br);
      for K in 1 .. Length(W) loop
         Br(W(K).I) := Br(W(K).I)+W(K).R*B(W(K).J);
      end loop;
      return Br;
   end "*";


   function "*"(B: Vector; W: Sparse) return Vector is
      Br: Vector(B'Range);
   begin
      SetZero(Br);
      for K in 1 .. Length(W) loop
         Br(W(K).J) := Br(W(K).J)+B(W(K).I)*W(K).R;
      end loop;
      return Br;
   end "*";


   function Id(N: Positive) return Matrix is
      pragma Suppress(Storage_Check);
      A: Matrix(1..N,1..N);
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            if I=J then A(I,J):=Comp_One;
            else  A(I,J):=Comp_Zero;
            end if;
         end loop;
      end loop;
      return A;
   end Id;


   --norms
   function Norm1(B: Vector) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+AbsVal(B(I));
      end loop;
      return R;
   end Norm1;


   function Norm1(A: Matrix) return Rep is
      R,Ri: Rep;
   begin
      R := Zero;
      for J in A'Range(2) loop
         Ri := Zero;
         for I in A'Range(1) loop
            Ri := Ri+AbsVal(A(I,J)); --sum over the column elemnts
         end loop;
         if Ri>R then
            R := Ri;
         end if;
      end loop;
      return R;
   end Norm1;


   function NormInf(A: Matrix) return Rep is
      R: Rep;
   begin
      R := Zero;
      for J in A'Range(2) loop
         for I in A'Range(1) loop
            if AbsVal(A(I,J))>R then
               R:=AbsVal(A(I,J));
            end if;
         end loop;
      end loop;
      return R;
   end NormInf;


   --regular procedures
   procedure Decomp(A: in out Matrix; Cond: out Rep; Pvt: out Pivot; Work: out Vector) is
      -- translated from numerical recipes
      N0,N,M: Integer;
      Anorm,Ynorm,Znorm: Rep;
      T: Comp;
   begin
      N0 := Work'First;
      N := Work'Last;

      Pvt(N) := N0;

      if N = N0 then
         Cond := One;
         if A(N0,N0) = Comp_Zero then
            Message("Lin.Ops.Decomp1",Matrix_Singular);
            return;
         end if;
         return;
      end if;
      Anorm := Zero;
      for J in N0 .. N loop
         T := Comp_Zero;
         for I in N0 .. N loop
            T := T+A(I,J);
         end loop;
         if AbsVal(T) > Anorm then
            Anorm := AbsVal(T);
         end if;
      end loop;
      for K in N0 .. N-1 loop
         M := K;
         for I in K+1 .. N loop
            if AbsVal(A(I,K)) > AbsVal(A(M,K)) then
               M := I;
            end if;
         end loop;
         Pvt(K) := M;
         if M /= K then
            Pvt(N) := -Pvt(N);
         end if;
         T := A(M,K);
         A(M,K) := A(K,K);
         A(K,K) := T;
         if T /= Comp_Zero then
            for I in K+1 .. N loop
               A(I,K) := Neg(A(I,K)/T);
            end loop;
            for J in K+1 .. N loop
               T := A(M,J);
               A(M,J) := A(K,J);
               A(K,J) := T;
               if T /= Comp_Zero then
                  for I in K+1 .. N loop
                     A(I,J) := A(I,J)+A(I,K)*T;
                  end loop;
               end if;
            end loop;
         end if;
      end loop;
      for K in N0 .. N loop
         T := Comp_Zero;
         for I in N0 .. K-1 loop
            T := T+A(I,K)*Work(I);
         end loop;
         if A(K,K) = Comp_Zero then
            Message("Lin.Ops.Decomp2",Matrix_Singular);
            return;
         end if;
         if RealPart(T) >=Zero then
            Work(K) := Neg(T+Comp_One)/A(K,K);
         else
            Work(K) := Neg(T-Comp_One)/A(K,K);
         end if;
      end loop;
      for K in reverse N0 .. N-1 loop
         T := Comp_Zero;
         for I in K+1 .. N loop
            T := T+A(I,K)*Work(K);
         end loop;
         Work(K) := T;
         M := Pvt(K);
         if M /= K then
            T := Work(M);
            Work(M) := Work(K);
            Work(K) := T;
         end if;
      end loop;
      Ynorm := Zero;
      for I in N0 .. N loop
         Ynorm := Ynorm+AbsVal(Work(I));
      end loop;
      Solve(A,Work,Pvt);
      Znorm := Zero;
      for I in N0 .. N loop
         Znorm := Znorm+AbsVal(Work(I));
      end loop;
      Cond := Anorm*Znorm/Ynorm;
      if Cond < One then
         Cond := One;
      end if;
   end Decomp;


   procedure Solve(A: in Matrix; B: in out Vector; Pvt: in Pivot) is
      --       use Complex_Types;
      -- solve A*B2=B1
      -- input:   A     triangularized matrix obtained from Decomp
      --          Pvt   pivot vector obtained from Decomp
      --          B     vector B1
      -- output:  B     vector B2
      N0,N,M: Integer;
      T: Comp;
   begin
      N0 := B'First;
      N := B'Last;
      for K in N0 .. N-1 loop
         M := Pvt(K);
         T := B(M);
         B(M) := B(K);
         B(K) := T;
         for I in K+1 .. N loop
            B(I) := B(I)+A(I,K)*T;
         end loop;
      end loop;
      for K in reverse N0+1 .. N loop
         B(K) := B(K)/A(K,K);
         T := Neg(B(K));
         for I in N0 .. K-1 loop
            B(I) := B(I)+A(I,K)*T;
         end loop;
      end loop;
      B(N0) := B(N0)/A(N0,N0);
   end Solve;


   procedure Solve(W: in Matrix; B: in Vector; X: in out Vector; Rsq: out Rep; Eps: in Rep := 1.0E-5) is
      -- translated procedure Sparse from numerical recipes
      Vsize: Integer;
      Gg,Gam,Dgg,Anum,Aden: Comp;
      Rp,Eps2,Bsq: Rep;
      G,H,Xi,Xj: Vector(B'Range);
   begin
      Vsize := B'Last-B'First+1;
      Eps2 := Rep(Vsize)*Eps*Eps;
      for Irst in 1 .. 3 loop    -- restart at most 3 times
         Xi := W*X;
         Rp := Zero;
         Bsq := Zero;
         for J in B'Range loop
            Bsq := Bsq+AbsVal(B(J)*B(J));
            Xi(J) := Xi(J)-B(J);
            Rp := Rp+AbsVal(Xi(J)*Xi(J));
         end loop;
         G := Xi*W;
         for J in B'Range loop
            G(J) := Neg(G(J));
            H(J) := G(J);
         end loop;
         for Chances in reverse 1 .. 100*Vsize loop
            Xi := W*H;
            Anum := Comp_Zero;
            Aden := Comp_Zero;
            for J in B'Range loop
               Anum := Anum+G(J)*H(J);
               Aden := Aden+Xi(J)*Xi(J);
            end loop;
            if Aden = Comp_Zero then
               Message("Lin.Ops.Solve",Matrix_Singular);
               return;
            end if;
            Anum := Anum/Aden;
            for J in B'Range loop
               Xi(J) := X(J);
               X(J) := X(J)+Anum*H(J);
            end loop;
            Xj := W*X;
            Rsq := Zero;
            for J in B'Range loop
               Xj(J) := Xj(J)-B(J);
               Rsq := Rsq+AbsVal(Xj(J)*Xj(J));
            end loop;
            if Rsq=Rp or Rsq<=Bsq*Eps2 then return; end if;
            if Rsq>Rp then
               X := Xi;
               exit when True;               --- restart
            end if;
            Rp := Rsq;
            Xi := Xj*W;
            Gg := Comp_Zero;
            Dgg := Comp_Zero;
            for J in B'Range loop
               Gg := Gg+G(J)*G(J);
               Dgg := Dgg+(Xi(J)+G(J))*Xi(J);
            end loop;
            if Gg=Comp_Zero then return; end if;
            Gam := Dgg/Gg;
            for J in B'Range loop
               G(J) := Neg(Xi(J));
               H(J) := G(J)+Gam*H(J);
            end loop;
            if Chances=1 then
               Put("Too many iterations in Solve");
               raise Getting_Tired;
            end if;
         end loop;
      end loop;
   end Solve;


   procedure Invert(A: in out Matrix; Output: in Boolean:=False) is
      pragma Suppress(Storage_Check);
      Pvt: Pivot(A'Range(1));
      BB,B: Vector(A'Range(1));
      AA,Aux: Matrix(A'Range(1),A'Range(2));
      Cond: Rep;
   begin
      AA:=A;
      Aux := A;
      Decomp(Aux,Cond,Pvt,B);
      for J in A'Range(2) loop
         SetZero(B);
         B(J) := Comp_One;
         BB:=B;
         Solve(Aux,B,Pvt);
         if Output then
            New_Line;
            Put("l_1 norm in Invert.Lin-Ops:");Txt_Put(Current_Output,Norm1(AA*B-BB));
         end if;
         SetColumn(J,B,A);
      end loop;
   end Invert;


   procedure ImproveInv(A: in  Matrix; InvA: in out Matrix; Err: in Rep := 1.0E-10; Output: in Boolean:=False) is
      Iter: constant Integer := 2000;
      Er: Rep;
      Iden: Matrix(A'Range(1),A'Range(2));
   begin
      if A'Length(1)/=A'Length(2)  then
         Message("ImproveInv.Lin-Ops",Index_Error);
      end if;

      SetZero(Iden);
      for I in Iden'Range(1) loop
         Iden(I,I):=Complecs(One,Zero);
      end loop;

      for I in 1 .. Iter loop
         InvA:=InvA*(A*InvA)-Complecs(Two,Zero)*InvA;
         Er:=Norm1(A*InvA-Iden);
         if Output then
            New_Line;
            Put("l_1 norm in ImproveInv.Lin-Ops: ");Txt_Put(Current_Output,Er);
         end if;
         if Er<Err then
            New_Line;
            Put("l_1 norm in ImproveInv.Lin-Ops: ");Txt_Put(Current_Output,Er);
         end if;
      end loop;
      Put(Current_Output,Iter,3,10);
      Put(" iterations in ImprobeInv.Lin-Ops");
      raise Getting_Tired;
   end ImproveInv;


   procedure ImproveSol(A: in  Matrix; B: in Vector; X: in out Vector; Err: in Rep := 1.0E-10; Output: in Boolean:=False) is
      Iter: constant Integer := 2000;
      Cond,Er: Rep;
      Pvt: Pivot(X'Range);
      Work,Del: Vector(X'Range);
      Aux: Matrix(A'Range(1),A'Range(2));
   begin

      if A'Length(1)/=A'Length(2)  then
         Message("ImproveInv.Lin-Ops",Index_Error);
      end if;
      Er:=Norm1(A*X-B);
      if Output then
         New_Line;
         Put("initial l_1 norm in ImproveSol.Lin-Ops:");Txt_Put(Current_Output,Er);
      end if;
      if Er<Err then
         return;
      end if;

      Aux:=A;
      Decomp(Aux,Cond,Pvt,Work);

      for I in 1 .. Iter loop
         Del:=B-A*X;
         Er:=Norm1(Del);
         Solve(Aux,Del,Pvt);
         X:=Del-X;
         if Output then
            New_Line;
            Put("l_1 norm in ImproveInv.Sol-Ops: ");Txt_Put(Current_Output,Er);
         end if;
         if Er<Err then
            New_Line;
            Put("l_1 norm in ImproveInv.Sol-Ops: ");Txt_Put(Current_Output,Er);
         end if;
      end loop;
      Put(Current_Output,Iter,3,10);
      Put(" iterations in ImproveSol.Lin-Ops");
      raise Getting_Tired;
   end ImproveSol;


   --sparse
   procedure AddComponent(U: in Trip; W: in out Sparse; Cut: in Rep := Zero) is
      L: Integer;
   begin
      if AbsVal(U.R)>Cut then
         L := Length(W)+1;
         if L>W'Last then
            Put("no more room in AddComponent");
            raise Matrix_Full;
         end if;
         W(L) := U;
         W(0).I := L;    -- new length of W
         W(0).J := 1;    -- may need cleanup
      end if;
   end AddComponent;


   function Length(W: in Sparse) return Integer is
   begin
      return W(0).I;
   end Length;


   procedure Solve(W: in Sparse; B: in Vector; X: in out Vector; Rsq: out Comp; Eps: in Rep := 1.0E-15) is
      -- translated procedure Sparse from numerical recipes
      Vsize: Integer;
      Rp,Gg,Gam,Dgg,Bsq,Anum,Aden: Comp;
      G,H,Xi,Xj: Vector(B'Range);
      Eps2: Rep;
   begin
      Vsize := B'Last-B'First+1;
      Eps2 := Rep(Vsize)*Eps*Eps;
      for Irst in 1 .. 3 loop    -- restart at most 3 times
         Xi := W*X;
         Rp := Comp_Zero;
         Bsq := Comp_Zero;
         for J in B'Range loop
            Bsq := Bsq+B(J)*B(J);
            Xi(J) := Xi(J)-B(J);
            Rp := Rp+Xi(J)*Xi(J);
         end loop;
         G := Xi*W;
         for J in B'Range loop
            G(J) := Neg(G(J));
            H(J) := G(J);
         end loop;
         for Chances in reverse 1 .. 100*Vsize loop
            Xi := W*H;
            Anum := Comp_Zero;
            Aden := Comp_Zero;
            for J in B'Range loop
               Anum := Anum+G(J)*H(J);
               Aden := Aden+Xi(J)*Xi(J);
            end loop;
            if Aden = Comp_Zero then
               Message("Lin.Ops.Solve",Matrix_Singular);
               return;
            end if;
            Anum := Anum/Aden;
            for J in B'Range loop
               Xi(J) := X(J);
               X(J) := X(J)+Anum*H(J);
            end loop;
            Xj := W*X;
            Rsq := Comp_Zero;
            for J in B'Range loop
               Xj(J) := Xj(J)-B(J);
               Rsq := Rsq+Xj(J)*Xj(J);
            end loop;
            if Rsq=Rp or AbsVal(Rsq)<=AbsVal(Bsq)*Eps2 then New_Line; Put("First scenario"); return; end if;
            --           if AbsVal(Rsq)>AbsVal(Rp) then
            --              X := Xi;
            --              New_Line;
            --              Put("Restart");
            --              exit when True;               --- restart
            --           end if;
            Rp := Rsq;
            Xi := Xj*W;
            Gg := Comp_Zero;
            Dgg := Comp_Zero;
            for J in B'Range loop
               Gg := Gg+G(J)*G(J);
               Dgg := Dgg+(Xi(J)+G(J))*Xi(J);
            end loop;
            if Gg=Comp_Zero then New_Line; Put("Second scenario"); return; end if;
            Gam := Dgg/Gg;
            for J in B'Range loop
               G(J) := Neg(Xi(J));
               H(J) := G(J)+Gam*H(J);
            end loop;
            if Chances=1 then
               Put("Too many iterations in Solve");
               raise Getting_Tired;
            end if;
         end loop;
      end loop;
   end Solve;


   --eigensolvers
   procedure Eigenvector(A: in Matrix; R: in Comp;  B: out Vector; Eps: in Rep := 1.0E-8) is
      Iter: constant Integer := 20000;
      C: Rep;
      Pvt: Pivot(B'Range);
      Bt: Vector(B'Range);
      Aux: Matrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      C := One/Sqrt(Rep(B'Last-B'First+1));
      for I in B'Range loop
         Aux(I,I) := Aux(I,I)-R;
         B(I) := Complecs(C,Zero);
      end loop;
      Decomp(Aux,C,Pvt,Bt);
      for I in 1 .. Iter loop
         Bt := B;
         Solve(Aux,B,Pvt);
         C := One/Norm2(B);
         if RealPart(B*Bt)<Zero then C := -C; end if;
         Mult(Complecs(C,Zero),B);
         Sub(B,Bt);
         New_Line;
         Txt_Put(Current_Output,Norm2(Bt));
         if Norm2(Bt)<Eps then return; end if;
      end loop;
      Put(Current_Output,Iter,3,10);
      Put(" iterations in Eigenvector");
      --  raise Getting_Tired;
   end Eigenvector;


   --conjugate methods
   procedure BCGstab(W: in Matrix; B: in Vector; X: in out Vector; Eps: in Rep := 1.0E-20) is
      I,Vsize: Integer;
      Rho,Rhold,Alpha,Beta,Gamma: Comp;
      Err: Rep;
      R,Rt,P,Q,T: Vector(B'Range);
   begin
      Put("Linear_Ops.BCGstab: ");
      Vsize := B'Last-B'First+1;
      R := B-W*X;
      Rt := R;
      I := 1;
      Rho := Comp_Zero;
      Alpha := Comp_Zero;
      Gamma := Comp_Zero;
      Q(1) := Comp_Zero;    -- just to avoid compiler warnings
      loop
         Rhold := Rho;
         Rho := (Rt*R);
         if I=1 then
            P := R;
         else
            Beta := (Rho/Rhold)*(Alpha/Gamma);
            for J in B'Range loop
               P(J) := R(J)+Beta*(P(J)-Gamma*Q(J));
            end loop;
         end if;
         Q := W*P;
         Alpha := Rho/(Rt*Q);
         for J in B'Range loop
            R(J) := R(J)-Alpha*Q(J);
         end loop;
         T := W*R;
         Gamma := (T*R)/(T*T);
         Err := Zero;
         for J in B'Range loop
            X(J) := X(J)-Alpha*P(J)-Gamma*R(J);
            R(J) := R(J)-Gamma*T(J);
            Err := Err+AbsVal(R(J));
         end loop;
         --         exit when I=Vsize or Err<Eps;
         New_Line;
         Txt_Put(Current_Output,Err);
         exit when Err<Eps;
         I := I+1;
      end loop;
      Put(I);
      Put(" iterations, Err =");
      Put(Current_Output,Err,5,4);
      New_Line;
   end BCGstab;


   --POINTWISE OPERATIONS
   function  PointsProd(M,H: in MatrixPtr) return MatrixPtr is
      MH: MatrixPtr:= new Matrix(H'Range(1),H'Range(2));
   begin
      SetZero(MH);
      -- Contribution of the components of H1 cross H2
      for I in MH'Range(1) loop
         for J in MH'Range(2) loop
            MH(I,J):=M(I,J)*H(I,J);
         end loop;
      end loop;
      return MH;
   end PointsProd;


   procedure  PointsProd(M: in MatrixPtr; H: in out MatrixPtr)  is
   begin
      for I in M'Range(1) loop
         for J in M'Range(2) loop
            H(I,J):=M(I,J)*H(I,J);
         end loop;
      end loop;
   end PointsProd;


   function  PointsProd(M,H: in Matrix) return Matrix is
      MH: Matrix(H'Range(1),H'Range(2));
   begin
      SetZero(MH);
      -- Contribution of the components of H1 cross H2
      for I in MH'Range(1) loop
         for J in MH'Range(2) loop
            MH(I,J):=M(I,J)*H(I,J);
         end loop;
      end loop;
      return MH;
   end PointsProd;


   procedure  PointsProd(M: in Matrix; H: in out Matrix)  is
   begin
      for I in M'Range(1) loop
         for J in M'Range(2) loop
            H(I,J):=M(I,J)*H(I,J);
         end loop;
      end loop;
   end PointsProd;


   procedure  PointsProdAdd(M: in Matrix; H: in out Matrix)  is
   begin
      for I in M'Range(1) loop
         for J in M'Range(2) loop
            H(I,J):=M(I,J)*H(I,J)+M(I,J);
         end loop;
      end loop;
   end PointsProdAdd;


   function  PointsQuot(M,H: in Matrix) return Matrix is
      MH: Matrix(H'Range(1),H'Range(2));
   begin
      SetZero(MH);
      -- Contribution of the components of H1 cross H2
      for I in MH'Range(1) loop
         for J in MH'Range(2) loop
            -- if AbsVal(H(I,J))>Rep(1.0E-5) and I/=H'First(1) then
            if  H(I,J)/=Comp_Zero  then
               MH(I,J):=M(I,J)/H(I,J);
            else
               MH(I,J):=M(I,J);
            end if;
         end loop;
      end loop;
      return MH;
   end PointsQuot;


   procedure  PointsQuot(M: in Matrix; H: in out Matrix)  is
   begin
      for I in M'Range(1) loop
         for J in M'Range(2) loop
            if  H(I,J)/=Comp_Zero  then
               H(I,J):=M(I,J)/H(I,J);
            else
               H(I,J):=M(I,J);
            end if;
         end loop;
      end loop;
   end PointsQuot;


   function  PointsProd(M,H: in RealMatrix) return RealMatrix is
      MH: RealMatrix(H'First(1)..H'Last(1),H'First(2)..H'Last(2));
   begin
      SetZero(MH);
      for I in MH'Range(1) loop
         for J in MH'Range(2) loop
            MH(I,J):=M(I,J)*H(I,J);
         end loop;
      end loop;
      return MH;
   end PointsProd;


   procedure  PointsProd(M: in RealMatrix; H: in out RealMatrix)  is
   begin
      for I in M'Range(1) loop
         for J in M'Range(2) loop
            H(I,J):=M(I,J)*H(I,J);
         end loop;
      end loop;
   end PointsProd;


   function  PointsQuot(M,H: in RealMatrix) return RealMatrix is
      MH: RealMatrix(H'First(1)..H'Last(1),H'First(2)..H'Last(2));
   begin
      SetZero(MH);
      for I in MH'Range(1) loop
         for J in MH'Range(2) loop
            if H(I,J)/=Zero then
               MH(I,J):=M(I,J)/H(I,J);
            else
               New_Line;
               Put("Zero value in division");
               MH(I,J):=Rep'Last;
            end if;
         end loop;
      end loop;
      return MH;
   end PointsQuot;


   procedure PointsQuot(M: in RealMatrix; H: in out RealMatrix)  is
   begin
      for I in H'Range(1) loop
         for J in H'Range(2) loop
            if H(I,J)/=Zero then
               H(I,J):=M(I,J)/H(I,J);
            else
               New_Line;
               Put("Zero value in division");
               H(I,J):=Rep'Last;
            end if;
         end loop;
      end loop;
   end PointsQuot;


   function Sum(A1,A2: in Matrix) return Matrix is
      A: Matrix(Integer'Min(A1'First(1),A2'First(1))..Integer'Max(A1'Last(1),A2'Last(1)),
                Integer'Min(A1'First(2),A2'First(2))..Integer'Max(A1'Last(2),A2'Last(2)));
   begin
      for I in  A'Range(1) loop
         for J in  A'Range(2) loop
            if InRange(I,J,A1) and not InRange(I,J,A2) then
               A(I,J):=A1(I,J);
            elsif not InRange(I,J,A1) and InRange(I,J,A2) then
               A(I,J):=A2(I,J);
            end if;
         end loop;
      end loop;

      for I in  Integer'Max(A1'First(1),A2'First(1)) .. Integer'Min(A1'Last(1),A2'Last(1)) loop
         for J in  Integer'Max(A1'First(2),A2'First(2)) .. Integer'Min(A1'Last(2),A2'Last(2))  loop
            A(I,J) := A1(I,J)+A2(I,J);
         end loop;
      end loop;
      return A;
   end Sum;


   function Diff(A1,A2: in Matrix) return Matrix is
      A: Matrix(Integer'Min(A1'First(1),A2'First(1))..Integer'Max(A1'Last(1),A2'Last(1)),
                Integer'Min(A1'First(2),A2'First(2))..Integer'Max(A1'Last(2),A2'Last(2)));
   begin
      for I in  A'Range(1) loop
         for J in  A'Range(2) loop
            if InRange(I,J,A1) and not InRange(I,J,A2) then
               A(I,J):=A1(I,J);
            elsif not InRange(I,J,A1) and InRange(I,J,A2) then
               A(I,J):=Neg(A2(I,J));
            end if;
         end loop;
      end loop;

      for I in  Integer'Max(A1'First(1),A2'First(1)) .. Integer'Min(A1'Last(1),A2'Last(1)) loop
         for J in  Integer'Max(A1'First(2),A2'First(2)) .. Integer'Min(A1'Last(2),A2'Last(2))  loop
            A(I,J) := A1(I,J)-A2(I,J);
         end loop;
      end loop;
      return A;
   end  Diff;


   function PointsConj(A: in Matrix) return Matrix is
      B: Matrix(A'Range(1),A'Range(2));
   begin
      SetZero(B);
      for I in B'Range(1) loop
         for J in B'Range(2) loop
            B(I,J):=Conj(A(I,J));
         end loop;
      end loop;
      return B;
   end PointsConj;


   function FouConj(A: in Matrix) return Matrix is
      B: Matrix(A'Range(1),-A'Last(2)..-A'First(2));
   begin
      SetZero(B);
      for I in B'Range(1) loop
         for J in B'Range(2) loop
            B(I,J):=Conj(A(I,-J));
         end loop;
      end loop;
      return B;
   end FouConj;


   procedure Prod(M,H: in Vector; MH: out Vector)  is
   begin
      -- Contribution of the components of H1 cross H2
      for I in MH'Range loop
         MH(I):=Comp_Zero;
         for K in M'Range loop
            if I-K>=H'First and I-K<=H'Last then
               MH(I):=MH(I)+M(K)*H(I-K);
            end if;
         end loop;
      end loop;
   end Prod;


   function Prod(M,H: in Vector) return Vector is
      MH: Vector(H'Range);
   begin
      -- Contribution of the components of H1 cross H2
      for I in MH'Range loop
         MH(I):=Comp_Zero;
         for K in M'Range loop
            if I-K>=H'First and I-K<=H'Last then
               MH(I):=MH(I)+M(K)*H(I-K);
            end if;
         end loop;
      end loop;
      return MH;
   end Prod;


   procedure Prod(M,H: in Matrix; MH: out Matrix)  is
   begin
      SetZero(MH);
      -- Contribution of the components of H1 cross H2
      for I in MH'Range(1) loop
         for J in MH'Range(2) loop
            for K in M'Range(2) loop
               if InRange(I,K,M) and InRange(I,J-K,H) then
                  MH(I,J):=MH(I,J)+M(I,K)*H(I,J-K);
               end if;
            end loop;
         end loop;
      end loop;
   end Prod;


   function  Prod(M,H: in Matrix) return Matrix is
      MH: Matrix(H'Range(1),H'Range(2));
   begin
      SetZero(MH);
      -- Contribution of the components of H1 cross H2
      for I in MH'Range(1) loop
         for J in MH'Range(2) loop
            for K in M'Range(2) loop
               if InRange(I,K,M) and InRange(I,J-K,H) then
                  MH(I,J):=MH(I,J)+M(I,K)*H(I,J-K);
               end if;
            end loop;
         end loop;
      end loop;
      return MH;
   end Prod;


   function NormInf(A: MatrixPtr) return Rep is
      R: Rep;
   begin
      R := Zero;
      for J in A'Range(2) loop
         for I in A'Range(1) loop
            if AbsVal(A(I,J))>R then
               R:=AbsVal(A(I,J));
            end if;
         end loop;
      end loop;
      return R;
   end NormInf;


   function "-"(A1,A2: in MatrixPtr) return MatrixPtr is
      A: MatrixPtr:= new Matrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
      return A;
   end "-";
end Lin.Ops;
