with Reps.IO,Comps.IO,Ada.Sequential_IO;
use Reps.IO,Comps.IO;
with Round_Up;

generic
package Lin.Ops is

   pragma Suppress(Storage_Check);

   package Sequential_Rep_IO is new Ada.Sequential_IO(Rep);
   package Sequential_Trip_IO is new Ada.Sequential_IO(Trip);

   type MatrixPtr is access Matrix;

   procedure SetZero(A: out MatrixPtr);
   procedure Add(B1: in RealVector; B2: in out RealVector);                           -- B2 := B2+B1
   procedure Add(A1: in RealMatrix; A2: in out RealMatrix);                           -- A2 := A2+A1
   procedure Add(R: in Rep; A: in out RealMatrix);                                    -- A := A+R*Identity
   procedure Sub(B1: in RealVector; B2: in out RealVector);                           -- B2 := B2-B1
   procedure Sub(A1: in RealMatrix; A2: in out RealMatrix);                           -- A2 := A2-A1
   procedure Mult(R: in Rep; B: in out RealVector);                                   -- B := R*B
   procedure Mult(R: in Rep; A: in out RealMatrix);                                   -- A := R*A
   procedure Mult(A1: in RealMatrix; A2: in out RealMatrix);                          -- A2 := A1*A2
   function  Prod(B0,B1,B2: RealVector) return Rep;
   function  RealId(N: Positive) return RealMatrix;
   procedure UnitVector(B: in out RealVector);

   function "-"(B: RealVector) return RealVector;                                     -- -B
   function "+"(B1,B2: RealVector) return RealVector;                                 -- B1+B2
   function "+"(A1,A2: in RealMatrix) return RealMatrix;                              -- A1+A2
   function "-"(B1,B2: RealVector) return RealVector;                                 -- B1-B2
   function "-"(A1,A2: in RealMatrix) return RealMatrix;                              -- A1-A2
   function "*"(B1,B2: RealVector) return Rep;                                        -- L2 inner product <B1,B2>
   function "*"(R: Rep; B: RealVector) return RealVector;                             -- R*B
   function "*"(R: Rep; A: RealMatrix) return RealMatrix;                             -- R*A
   function "*"(B1,B2: RealVector) return RealVector;                                 -- Diag(B1)*B2
   function "*"(A: RealMatrix; B: RealVector) return RealVector;                      -- A*B
   function "*"(B: RealVector; A: RealMatrix) return RealVector;                      -- Transpose(A)*B
   function "*"(A1,A2: RealMatrix) return RealMatrix;                                 -- A1*A2

   function Norm1(A: RealMatrix) return Rep;
   function Norm1(B: RealVector) return Rep;
   function Norm2(B: RealVector) return Rep;

   procedure Decomp(A: in out RealMatrix; Cond: out Rep; Pvt: out Pivot; Work: out RealVector); -- Triangularize A, LU decomposition
   procedure Solve(A: in RealMatrix; B: in out RealVector; Pvt: in Pivot);
   procedure Solve(W: in RealMatrix; B: in RealVector; X: in out RealVector; Rsq: out Rep; Eps: in Rep := 1.0E-5);
   procedure Invert(A: in out RealMatrix; Output: in Boolean:=False);
   procedure ImproveInv(A: in  RealMatrix; InvA: in out RealMatrix; Err: in Rep := 1.0E-5; Output: in Boolean:=False);

   procedure Balanc(A: in out RealMatrix);                                                       -- Balance A, preserves spectrum
   procedure ElmHes(A: in out RealMatrix);                                                       -- Reduce A to Hessenberg form
   procedure Hqr(A: in out RealMatrix; Wr,Wi: out RealVector);                                   -- Find Eigenvalues of Hessenberg matrix
   procedure Eigenvector(A: in RealMatrix; R: in Rep; B: in out RealVector; Eps: in Rep := 1.0E-8);
   procedure Eigenvalues(A: in RealMatrix; Br,Bi: out RealVector);                               -- Eigenvalues of A (real and imag parts)

   procedure Descent(W: in RealMatrix; B: in RealVector; X: in out RealVector; Eps: in Rep := 1.0E-8);
   procedure BCG(W: in RealMatrix; B: in RealVector; X: in out RealVector; Eps: in Rep := 1.0E-8);
   procedure BCG(W: in RealMatrix; Bf,B: in RealVector; X: in out RealVector; Eabs,Erel: in Rep);
   procedure BCGstab(W: in RealMatrix; B: in RealVector; X: in out RealVector; Eps: in Rep := 1.0E-8);
   procedure BCGstab(W: in RealMatrix; Bf,B: in RealVector; X: in out RealVector; Eabs,Erel: in Rep);

   procedure SVD( A: in RealMatrix;U: out RealMatrix; W: out RealMatrix; V: out RealMatrix);
   procedure SVDSolve(A: in  RealMatrix; X: in out RealVector; Output: in Boolean:=False);
   procedure InvertSVD(A: in out RealMatrix;Output: in Boolean:=False);

   procedure Add(B1: in Vector; B2: in out Vector);                           -- B2 := B2+B1
   procedure Add(A1: in Matrix; A2: in out Matrix);                           -- A2 := A2+A1
   procedure Add(R: in Comp; A: in out Matrix);                               -- A := A+R*Identity
   procedure Sub(B1: in Vector; B2: in out Vector);                           -- B2 := B2-B1
   procedure Sub(A1: in Matrix; A2: in out Matrix);                           -- A2 := A2-A1
   procedure Mult(R: in Comp; B: in out Vector);                              -- B := R*B
   procedure Mult(R: in Comp; A: in out Matrix);                              -- A := R*A
   procedure Mult(A1: in Matrix; A2: in out Matrix);                          -- A2 := A1*A2
   function Less(L,R,Base: in Comp) return Boolean;
   function Id(N: Positive) return Matrix;
   procedure Swap(X,Y: in out Comp);
   procedure Sift_down(A : in out Vector;From: Positive; Base: in Comp);
   procedure Sift_Up(A : in out Vector; Base: in Comp);
   procedure Heap_Sort(A : in out Vector; Base: in Comp);

   function "-"(B: Vector) return Vector;                                     -- -B
   function "+"(B1,B2: Vector) return Vector;                                 -- B1+B2
   function "+"(A1,A2: in Matrix) return Matrix;                              -- A1+A2
   function "-"(B1,B2: Vector) return Vector;                                 -- B1-B2
   function "-"(A1,A2: in Matrix) return Matrix;                              -- A2-A1
   function "*"(B1,B2: Vector) return Comp;                                   -- L2 inner product <B1,B2>
   function "*"(R: Comp; B: Vector) return Vector;                            -- R*B
   function "*"(R: Comp; A: Matrix) return Matrix;                            -- R*A
   function "*"(B1,B2: Vector) return Vector;                                 -- Diag(B1)*B2
   function "*"(A: Matrix; B: Vector) return Vector;                          -- A*B
   function "*"(B: Vector; A: Matrix) return Vector;                          -- Transpose(A)*B
   function "*"(A1,A2: Matrix) return Matrix;                                 -- A1*A2
   function "*"(W: Sparse; B: Vector) return Vector;                          -- W*B
   function "*"(B: Vector; W: Sparse) return Vector;                          -- Transpose(W)*B

   function Norm1(A: Matrix) return Rep;
   function NormInf(A: Matrix) return Rep;
   function Norm1(B: Vector) return Rep;
   function Norm2(B: Vector) return Rep;
   procedure Decomp(A: in out Matrix; Cond: out Rep; Pvt: out Pivot; Work: out Vector);         -- Triangularize A, LU decomposition
   procedure Solve(A: in Matrix; B: in out Vector; Pvt: in Pivot);
   procedure Solve(W: in Matrix; B: in Vector; X: in out Vector; Rsq: out Rep; Eps: in Rep := 1.0E-5);
   procedure Invert(A: in out Matrix; Output: in Boolean:=False);
   procedure ImproveInv(A: in  Matrix; InvA: in out Matrix; Err: in Rep := 1.0E-10; Output: in Boolean:=False);
   procedure ImproveSol(A: in  Matrix; B: in Vector; X: in out Vector; Err: in Rep := 1.0E-10; Output: in Boolean:=False);
   procedure AddComponent(U: in Trip; W : in out Sparse; Cut: in Rep := Zero);                   -- Add R to W(i,j) if >Cut
   function Length(W: in Sparse) return Integer;                                                 -- Number of filled (nonzero) elements
   procedure Solve(W: in Sparse; B: in Vector; X: in out Vector; Rsq: out Comp; Eps: in Rep := 1.0E-15);    -- Solve W*X=B (steepest descent)
   procedure Eigenvector(A: in Matrix; R: in Comp;  B: out Vector; Eps: in Rep := 1.0E-8);       -- Try to solve (A-R)*B=0
   procedure BCGstab(W: in Matrix; B: in Vector; X: in out Vector; Eps: in Rep := 1.0E-20);

   --Ponters
   procedure Add(A1: in MatrixPtr; A2: in out MatrixPtr);
   function "-"(A1,A2: in MatrixPtr) return MatrixPtr;
   function NormInf(A: MatrixPtr) return Rep;

   procedure Prod(M,H: in Matrix; MH: out Matrix);
   function  Prod(M,H: in Matrix) return Matrix;
   function PointsProd(M,H: in MatrixPtr) return MatrixPtr;
   procedure PointsProd(M: in MatrixPtr; H: in out MatrixPtr);

   function PointsProd(M,H: in Matrix) return Matrix;
   procedure PointsProd(M: in Matrix; H: in out Matrix);
   procedure  PointsProdAdd(M: in Matrix; H: in out Matrix);

   function PointsQuot(M,H: in Matrix) return Matrix;
   procedure PointsQuot(M: in Matrix; H: in out Matrix);
   function PointsProd(M,H: in RealMatrix) return RealMatrix;
   procedure PointsProd(M: in RealMatrix; H: in out RealMatrix);
   function PointsQuot(M,H: in RealMatrix) return RealMatrix;
   procedure PointsQuot(M: in RealMatrix; H: in out RealMatrix);
   procedure Prod(M,H: in Vector; MH: out Vector);
   function Prod(M,H: in Vector) return Vector;
   function Sum(A1,A2: in Matrix) return Matrix;
   function Diff(A1,A2: in Matrix) return Matrix;
   function PointsConj(A: in Matrix) return Matrix;
   function FouConj(A: in Matrix) return Matrix;
end Lin.Ops;
